import numpy as np
import pandas as pd
from pandas.io import sql
import sqlite3
from sqlalchemy import create_engine
from sqlalchemy import text
import astropy.units as u
from numpy.core.records import fromarrays
from scipy.io import savemat
__author__ = 'jundurraga-ucl'


class SubjectInformation(object):
    def __init__(self, subject_id: str = '',
                 gender: str = '',
                 age: float = None,
                 date_of_birth: str = '',
                 handedness: str = '',
                 comments: str = '',
                 group: str = ''):
        self.subject_id = subject_id
        self.gender = gender
        self.age = age
        self.date_of_birth = date_of_birth
        self.handedness = handedness
        self.comments = comments
        self.group = group


class MeasurementInformation(object):
    def __init__(self, experiment: str = '', condition: str = '', date: str = '', comments: str = ''):
        self.experiment = experiment
        self.condition = condition
        self.date = date
        self.comments = comments


class PandasDataTable(object):
    def __init__(self, table_name: str = '', pandas_df: pd.DataFrame = None):
        self.table_name = table_name
        self.pandas_df = pandas_df


def pandas_with_units_to_sql_compatible(data_frame: pd.DataFrame = None):
    """
    Convert pandas data frame with astropy units to pandas compatible with sql
    :param data_frame:
    :return: data frame with columns where quantities and units are in separate columns
    """
    for column in data_frame:
        _condition_1 = np.any(np.array([isinstance(_v, u.quantity.Quantity)
                                        for _v in data_frame[column].to_numpy()]))
        _condition_2 = False
        if np.any(_condition_1):
            _condition_2 = np.any(np.array([_v.unit != u.dimensionless_unscaled
                                            for _v in data_frame[column].to_numpy()
                                            if isinstance(_v, u.Quantity)]))
        if _condition_2:
            unit_column = column + '_unit'
            data_frame[unit_column] = None
            for _i, _v in enumerate(data_frame[column].to_numpy()):
                if isinstance(_v, u.Quantity):
                    data_frame.iloc[_i, data_frame.columns.get_loc(unit_column)] = _v.unit.to_string()
                else:
                    data_frame.iloc[_i, data_frame.columns.get_loc(unit_column)] = u.dimensionless_unscaled.to_string()
        if _condition_1:
            for _i, _v in enumerate(data_frame[column].to_numpy()):
                if isinstance(_v, u.Quantity):
                    data_frame.iloc[_i, data_frame.columns.get_loc(column)] = _v.value
                else:
                    data_frame.iloc[_i, data_frame.columns.get_loc(column)] = _v

    return data_frame


def store_data(database_path='',
               subject_info: SubjectInformation = None,
               measurement_info: MeasurementInformation = MeasurementInformation(),
               stimuli_info: dict = None,
               recording_info: dict = None,
               pandas_df: [PandasDataTable] = None,
               check_duplicates: bool = True
               ):
    # connect and add items to data base
    # db = sqlite3.connect(database_path)
    engine = create_engine("sqlite:///" + database_path)
    cnn = engine.connect()
    # cursor = db.cursor()
    statement = "SELECT name FROM sqlite_master WHERE type='table';"
    print('Storing data in %s' % database_path)
    all_tables = cnn.execute(text(statement)).fetchall()
    if ('subjects',) in all_tables:
        cursor = engine.connect().execute(text('select id from subjects ;'))
        _index = len(cursor.fetchall()) + 1
    else:
        _index = 1
    _new_df = dict({'anonymous_name': 'S' + str(_index)}, **subject_info.__dict__)

    _id_subject = upsert_db(table_name='subjects', new_df=pd.DataFrame([_new_df]), db=engine,
                            column_names=['subject_id'], replace=False)

    # add measurement info
    _row_measurement_info = dict({'id_subject': _id_subject},
                                 **measurement_info.__dict__)
    _row_measurement_info = pandas_with_units_to_sql_compatible(pd.DataFrame([_row_measurement_info]))
    _id_measurement = upsert_db(table_name='measurement_info', new_df=_row_measurement_info,
                                db=engine,
                                column_names=list(_row_measurement_info.keys()), replace=False)

    # convert lists to string to be compatible with sqlite
    _stimuli_info = stimuli_info.copy()
    [_stimuli_info.update({_key: str(_value)}) for _key, _value in _stimuli_info.items() if isinstance(_value, list)]
    # add stimuli to table
    _row_stimuli = dict({'id_subject': _id_subject, 'id_measurement': _id_measurement}, **_stimuli_info)
    # check no dict is passed
    [_row_stimuli.pop(_key) for _key in list(_row_stimuli.keys()) if isinstance(_row_stimuli[_key], dict)]
    _row_stimuli = pandas_with_units_to_sql_compatible(pd.DataFrame([_row_stimuli]))
    _id_stimuli = upsert_db(table_name='stimuli', new_df=_row_stimuli, db=engine,
                            column_names=list(_row_stimuli.keys()), replace=False)

    # add recording processing to table
    _recording_info = recording_info.copy()
    [_recording_info.update({_key: str(_value)}) for
     _key, _value in _recording_info.items() if isinstance(_value, list)]
    _row_recording = dict({'id_stimuli': _id_stimuli, 'id_subject': _id_subject, 'id_measurement': _id_measurement},
                          **_recording_info)
    _row_recording = pandas_with_units_to_sql_compatible(pd.DataFrame([_row_recording]))
    _id_recording = upsert_db(table_name='recording', new_df=_row_recording, db=engine,
                              column_names=list(_row_recording.keys()), replace=False)

    for _df in pandas_df:
        _df.pandas_df = pandas_with_units_to_sql_compatible(_df.pandas_df)
        _df.pandas_df['id_subject'] = _id_subject
        _df.pandas_df['id_measurement'] = _id_measurement
        _df.pandas_df['id_stimuli'] = _id_stimuli
        _df.pandas_df['id_recording'] = _id_recording
        upsert_db(table_name=_df.table_name, new_df=_df.pandas_df, db=engine,
                  column_names=['id_subject', 'id_measurement', 'id_stimuli', 'id_recording'],
                  replace=check_duplicates,
                  check_duplicates=check_duplicates)
    cnn.close()
    engine.dispose()


def write_set(filename, processname, pipe_line, event_code, ref_el):
    """Export OutputNodedata to EEGLAB .set file.
    Exports either continuous EEG data or epoched EEG data from the pipe line named 'pipe_line' and the
    pipe line step named 'processname' to a .set file named 'filename'. Also exports the known reference electrode
    'ref_el', and events coded with 'event_code'.
    Export functionality was assessed by loading the exported data into eeglab2021.1.
    Code ported from MNE-Lab, https://github.com/cbrnr/mnelab/blob/main/mnelab/io/writers.py
    Caveat: EEG layout parameters are currently not converted, only the channel names are used! Also it is assumed,
    that the pipeline has been called with the run() method before saving the data up to the requested process.
    Some of the following code parts are creating "dummy" variables, to match specific variables needed in EEGlab,
    see https://eeglab.org/tutorials/ConceptsGuide/Data_Structures.html for EEGlab datastructure documentation.

    Input parameters:
    filename: full filepath + Name of file to save, e.g. /home/user/eeg/mydata.set
    processname: name of output node processing step you want to save
    pipe_line: pipe object
    event_code: event code associated with the eeg data. Events needs to have a code, time and duration in order to get
    exported. Duration of trigger for now is fixed to 5ms. Currently only tested with one single event code.
    ref_el: reference electrode

    Output parameters:
    None, will print a line to STDOUT if export successful, containing processname and filename.

    Example call:
    .....
    ref_el = 'Cz'
    eventcode = 241 #Trigger event code
    pipe_line.append(AutoRemoveBadChannels(pipe_line.get_process('referenced'),
                                            keep_input_node=False),
                      name='channel_cleaned')

    pipe_line.run()
    write_set('/channel_cleaned.set', 'channel_cleaned', pipe_line, eventcode, ref_el)
    """
    processed_data = pipe_line.get_process(processname)
    fs = processed_data.output_node.fs.value
    data = processed_data.output_node.get_data().value.astype('float64')
    times = processed_data.output_node.x.value
    if len(data.shape) == 3:  # for epoched data
        data = np.transpose(data, (1, 0, 2))  # EEGlab data format is channel x amplitude x epoch
        epochs = data.shape[2]
        epoch_metadata = fromarrays([
                                    np.arange(1, epochs+1),
                                    np.tile('square', epochs),
                                    np.arange(1, epochs+1),
                                    np.tile(0.005, epochs),
                                    np.tile(0.0, epochs)],
                                    names=["event", "eventtype", "eventposition", "eventduration", "eventlatency"])
        # create dummy event data for EEGlab/brainstorms events structures
        events = fromarrays([np.tile(event_code, epochs),
                             # calculates epoch start latency in samples
                             [max(times)-min(times) * i * fs for i in np.arange(1, epochs+1)],
                             np.tile((0.005 * fs), epochs),  # assume 5ms duration of epoch trigger
                             np.arange(1, epochs+1).astype('float64')],
                            names=["type", "latency", "duration", "epoch"])
    elif len(data.shape) == 2:  # for continuous data
        data = np.transpose(data, (1, 0))  # EEGlab data format is channel x amplitude
        epochs = 1
        epoch_metadata = 0
        events = fromarrays([np.squeeze(processed_data.output_node.events.get_events_code(event_code)),
                             processed_data.output_node.events.get_events_time(event_code).value * fs + 1.0,
                             # event time starting points are stored as sample index in EEGlab events structure
                             processed_data.output_node.events.get_events_duration(event_code)],
                            names=["type", "latency", "duration"])
    else:
        raise Exception('Data needs to have at least two dimensions (channel x amplitude)')

    ch_names = [i.label for i in processed_data.output_node.layout]
    chanlocs = fromarrays([ch_names], names=["labels"])

    savemat(filename, dict(EEG=dict(data=data,
                                    setname=filename,
                                    nbchan=data.shape[0],
                                    pnts=float(data.shape[1]),
                                    trials=epochs,
                                    srate=fs,
                                    ref=ref_el,
                                    xmin=times[0],
                                    xmax=times[-1],
                                    chanlocs=chanlocs,
                                    epoch=epoch_metadata,
                                    event=events,
                                    icawinv=[],
                                    icasphere=[],
                                    icaweights=[])),
            appendmat=False)
    print("Successfully saved " + processname + " to " + filename)


def find_item_in_table(table_name='', new_df=pd.DataFrame, db=sqlite3.Connection):
    cursor = db.cursor()
    statement = "SELECT name FROM sqlite_master WHERE type='table';"
    all_tables = cursor.execute(statement).fetchall()
    if (table_name,) in all_tables:
        table = sql.read_sql(text('select * from ' + table_name), db.connect())
        _found = table[new_df.columns].isin(new_df)
    return _found


def upsert_db(table_name='',
              new_df=pd.DataFrame,
              column_names: [str] = None,
              db=sqlite3.Connection,
              replace=True,
              check_duplicates=True
              ):
    """
   This function will upsert data to database therefore preventing duplicated rows.
   :param table_name: Name of the table where we want to add new data
   :param new_df: pandas data frame containing data to be appended
   :param column_names: list of columns in the database that should not be duplicated
   :param db: database connection
   :param replace: if true, it will always replace an existing entry, if false and a data entry exists then the index
   of this entry will be returned.
   :param check_duplicates: if true, it will check whether a similar row exits, it will find it and return the index.
   If false, it will always add a new row, regardless of whether a similar row exists.
   :return: indexes of new appended data
    """

    if not new_df.size:
        return None
    statement = text("SELECT name FROM sqlite_master WHERE type='table';")
    all_tables = db.connect().execute(statement).fetchall()
    connection = db.raw_connection()
    cursor = connection.cursor()
    add_new_rows = not check_duplicates
    if (table_name,) in all_tables:
        if check_duplicates:
            # find which rows already exists
            _condition = ''
            for _idx_col, _col in enumerate(column_names):
                sqlist = df_values_to_sqlist(new_df, _col)
                sep_and = '' if _idx_col == len(column_names) - 1 else ' and '
                _operator = 'is' if sqlist == '(NULL)' else 'in'
                _condition = _condition + '{:s} {:} {:} {:} '.format(''.join(_col), _operator, sqlist, sep_and)
            _table = cursor.execute('select id from ' + table_name + ' where ' + _condition + ' ;')
            _index = np.array(_table.fetchall())
            add_new_rows = _index.size == 0
            if _index.size and replace:
                # delete those rows that we are going to "upsert"
                _to_delete = array_values_to_sqlist(_index)
                _table = cursor.execute('delete from ' + table_name + ' where id in {:};'.format(_to_delete))
                connection.commit()
        # add data new data ensuring new indexes
        if add_new_rows or replace:
            _table = cursor.execute('select id from ' + table_name + ' ;')
            _all_ids = _table.fetchall()
            idx_off_set = np.max(_all_ids) + 1 if _all_ids else 0
            _index = new_df.index.values + idx_off_set
            new_df['id'] = _index
            new_df.to_sql(table_name, db, if_exists='append', index=False, chunksize=1000, method='multi')
    else:
        new_df.to_sql(table_name, db, if_exists='append', index=True, index_label='id', chunksize=1000, method='multi')
        # get new index from DB
        _id = sql.read_sql(text('select id from ' + table_name), db.connect())
        _index = _id.values
    # delete temp table
    connection.commit()
    cursor.close()
    return int(_index) if _index.size == 1 else _index


def df_values_to_sqlist(new_df: pd.DataFrame = None,
                        column_name: str = None):
    new_df_values = new_df[column_name].values
    sql_list_values = ''
    for idx_v, _v in enumerate(new_df_values):
        sep = '' if idx_v == len(new_df_values) - 1 else ','
        if isinstance(_v, str):
            sql_list_values = sql_list_values + '"{:}"{:}'.format(_v, sep)
        elif isinstance(_v, np.ndarray):
            if _v.size:
                sql_list_values = sql_list_values + '"{:}"{:}'.format(_v, sep)
            else:
                sql_list_values = sql_list_values + '{:}{:}'.format('NULL', sep)
        elif _v is None:
            sql_list_values = sql_list_values + '{:}{:}'.format('NULL', sep)
        elif _v == np.inf:
            sql_list_values = sql_list_values + '"{:}"{:}'.format('Inf', sep)
        elif _v == -np.inf:
            sql_list_values = sql_list_values + '"{:}"{:}'.format('-Inf', sep)
        else:
            sql_list_values = sql_list_values + '{:}{:}'.format(_v, sep)
    sql_list_values = '({:})'.format(sql_list_values)
    return sql_list_values


def array_values_to_sqlist(values: np.array = None):
    sql_list_values = ''
    for idx_v, _v in enumerate(values):
        _v = _v.squeeze()
        sep = '' if idx_v == len(values) - 1 else ','
        if isinstance(_v, str):
            sql_list_values = sql_list_values + '"{:}"{:}'.format(_v, sep)
        elif _v is None:
            sql_list_values = sql_list_values + '{:}{:}'.format('NULL', sep)
        elif _v == np.inf:
            sql_list_values = sql_list_values + '"{:}"{:}'.format('Inf', sep)
        elif _v == -np.inf:
            sql_list_values = sql_list_values + '"{:}"{:}'.format('-Inf', sep)
        else:
            sql_list_values = sql_list_values + '"{:}"{:}'.format(_v, sep)
    sql_list_values = '({:})'.format(sql_list_values)
    return sql_list_values



import numpy as np
import inspect
from peegy.definitions.channel_definitions import Domain
from peegy.definitions.eeg_definitions import EegPeak
from peegy.processing.tools.detection.definitions import PeakToPeakMeasure, PeakToPeakQuantity, TimePeakWindow, \
    PeaksContainer, PeakToPeakAmplitudeContainer
from peegy.processing.pipe.definitions import DataNode
import astropy.units as u


def get_peaks_2(time_vector=np.array([]), data_channels=np.array([]), rn=np.array([]), **kwargs):
    peak_windows = kwargs.get('peak_windows', np.array([]))
    channel_labels = kwargs.get('channel_labels', np.array([]))
    # classify all peaks
    peaks = detect_peaks(time_vector=time_vector,
                         data_channels=data_channels,
                         rn=rn,
                         peak_windows=peak_windows,
                         channel_labels=channel_labels)
    return peaks


def detect_peaks(peak_windows=np.array([TimePeakWindow]),
                 data_channels=np.array([]),
                 time_vector=np.array([]),
                 rn=np.array([]),
                 channel_labels=np.array([])):
    _detected_peaks = np.array([])
    if not channel_labels.size:
        channel_labels = np.arange(data_channels.size).astype(np.str)
    _peak_windows = sort_peak_windows(peak_windows)

    for i, _ch_label in zip(np.arange(data_channels.shape[1]), channel_labels):
        _peak_container = PeaksContainer(channel_label=_ch_label)
        for _t_w in _peak_windows:
            # apply time reference to current peal window (ini or end)
            if _t_w.target_channels is not None and _ch_label not in _t_w.target_channels:
                continue
            if _t_w.exclude_channels is not None and _ch_label in _t_w.exclude_channels:
                continue
            _t_w = update_time_window(peak_window=_t_w, peaks_container=_peak_container)
            _peak = find_time_peak(peak_time_window=_t_w,
                                   data=data_channels[:, i],
                                   rn=rn[i],
                                   time_vector=time_vector,
                                   channel_label=_ch_label
                                   )
            _peak_container.append(_peak)
        _detected_peaks = np.append(_detected_peaks, _peak_container)
    return _detected_peaks


def update_time_window(peak_window: TimePeakWindow, peaks_container: PeaksContainer = None):
    """
    This function search and apply reference time to passed windows. If a given ini_ref or end_ref was given to
    peak_window, we search the referential peak across all passed peaks within peaks_container.
    :param peak_window: TimePeakWindow class with searching parameters for a given peak
    :param peaks_container: PeaksContainer class with found peaks
    :return: updated version of peak_window (ini_time and/or end_time) to perform search
    """
    for _peak in peaks_container.get_peaks():
        if peak_window.ini_ref is not None and peak_window.ini_ref == _peak.peak_label:
            peak_window.ini_time = _peak.x
        if peak_window.end_ref is not None and peak_window.end_ref == _peak.peak_label:
            peak_window.end_time = _peak.x
    return peak_window


def label_peak(peak=EegPeak, time_window=TimePeakWindow):
    if peak is not None:
        peak.label_peak = time_window.label
        peak.show_label = time_window.show_label
        # make sure that there is not another peak having this label
        # for _p in all_peaks:
        #     if _p.label_peak == _detected_peak.label_peak and _detected_peak != _p:
        #         _p._detected_peak = None
        #         _p.show_label = False


def find_time_peak(peak_time_window: TimePeakWindow = None,
                   data=np.array([]),
                   time_vector=np.array([]),
                   rn: float = None,
                   channel_label=None,
                   ):
    _ini = np.searchsorted(time_vector, peak_time_window.ini_time)
    _end = np.searchsorted(time_vector, peak_time_window.end_time)
    _end = _end + 1 if _end == _ini else _end

    if peak_time_window.positive_peak:
        _idx = np.argmax(data[_ini:_end])
    else:
        _idx = np.argmin(data[_ini:_end])

    _idx += _ini
    _amp = data[_idx]
    _significant = np.abs(_amp) > 2.0 * rn
    snr = np.abs(_amp) / rn
    _val = {'x': time_vector[_idx],
            'rn': rn,
            'amp': _amp,
            'amp_snr': snr,
            'significant': _significant,
            'peak_label': peak_time_window.label,
            'show_label': peak_time_window.show_label,
            'show_peak': peak_time_window.show_peak,
            'positive': peak_time_window.positive_peak,
            'domain': Domain.time,
            'channel': channel_label}
    return EegPeak(**_val)


def measure_peak_to_peak_amplitudes(detected_peaks: np.array([PeaksContainer]) = None,
                                    peak_to_peak_amp_labels=np.array([PeakToPeakMeasure()])):
    if peak_to_peak_amp_labels is None:
        return
    amplitudes = np.array([])
    for peaks_container in detected_peaks:
        peak_to_peak_container = PeakToPeakAmplitudeContainer(channel_label=peaks_container.channel_label)
        for p_p_labels in peak_to_peak_amp_labels:
            ini_peak = [_peak for _peak in peaks_container.peaks if _peak.peak_label == p_p_labels.ini_peak]
            end_peak = [_peak for _peak in peaks_container.peaks if _peak.peak_label == p_p_labels.end_peak]
            _current_amp = {}
            if ini_peak and end_peak:
                amp = np.abs(ini_peak[0].amp - end_peak[0].amp)
                amp_sigma = (ini_peak[0].rn ** 2.0 + end_peak[0].rn ** 2.0) ** 0.5
                if ini_peak[0].amp >= end_peak[0].amp:
                    max_time = ini_peak[0].x
                else:
                    max_time = end_peak[0].x

                if ini_peak[0].amp < end_peak[0].amp:
                    min_time = ini_peak[0].x
                else:
                    min_time = end_peak[0].x
                _current_amp = {'max_time': max_time,
                                'min_time': min_time,
                                'ini_time': ini_peak[0].x,
                                'end_time': end_peak[0].x,
                                'rn': amp_sigma,
                                'amp': amp,
                                'amp_snr': np.inf * u.dimensionless_unscaled if amp_sigma == 0 else
                                amp / amp_sigma,
                                'amp_label': ini_peak[0].peak_label + '-' + end_peak[0].peak_label,
                                'channel': peaks_container.channel_label
                                }
                peak_to_peak_container.append(PeakToPeakQuantity(**_current_amp))
        amplitudes = np.append(amplitudes, peak_to_peak_container)
    return amplitudes


def sort_peak_windows(peak_windows=np.array([TimePeakWindow()])):
    for _i, _w in enumerate(peak_windows):
        _w_dependency_labels = [_w.ini_ref, _w.end_ref, _w.global_ref]
        for _dep in _w_dependency_labels:
            if _dep is not None:
                _idx = [_ip for _ip, _p in enumerate(peak_windows) if _p.label == _dep]
                if _idx:
                    # when find dependency, move it so dependency is first
                    if _idx[0] > _i:
                        peak_windows[_idx[0]], peak_windows[_i] = peak_windows[_i], peak_windows[_idx[0]]
                        # print [_p.label for _p in  peak_windows]
    return peak_windows


def join_to_channel_dict(dict_to_join=[{}], channels=[{}]):
    out = []
    for i, _ch in enumerate(channels):
        for _items in dict_to_join[i]:
            if inspect.isclass(type(_items)):
                _to_join = _items.__dict__
            else:
                _to_join = _items
            out.append(dict(_ch, **_to_join))
    return out


def join_arr_dict_to_arr_dict(arr_1=[{}], arr_2=[{}]):
    out = []
    for i, (_ch, _ndi) in enumerate(zip(arr_1, arr_2)):
        out.append(dict(_ch, **_ndi))
    return out


def detect_peaks_and_amplitudes(data_node=DataNode(),
                                time_peak_windows=np.array([TimePeakWindow()]),
                                eeg_peak_to_peak_measures=np.array([PeakToPeakMeasure()]),
                                ):
    peaks = None
    amplitudes = None
    if time_peak_windows is not None:
        # find peaks for all channels
        peaks = detect_peaks(time_vector=data_node.x,
                             data_channels=data_node.data,
                             channel_labels=np.array([_ch.label for _ch in data_node.layout]),
                             rn=data_node.rn,
                             peak_windows=time_peak_windows)
        if eeg_peak_to_peak_measures is not None:
            amplitudes = measure_peak_to_peak_amplitudes(detected_peaks=peaks,
                                                         peak_to_peak_amp_labels=eeg_peak_to_peak_measures)

    return peaks, amplitudes


def get_channel_peaks_and_windows(eeg_peaks=np.array([EegPeak]), eeg_time_windows=np.array([TimePeakWindow()]),
                                  channel_label=None):
    sub_peaks = np.array([])
    sub_windows = np.array([])
    if eeg_peaks is not None and len(eeg_peaks) > 0:
        sub_peaks = eeg_peaks.loc[eeg_peaks['channel'] == channel_label]
        for _w in eeg_time_windows:
            if _w.channel == channel_label:
                sub_windows = np.append(sub_windows, _w)
    return sub_peaks, sub_windows



import numpy as np
from peegy.processing.tools.filters.resampling import eeg_resampling
import matplotlib.pyplot as plt
import astropy.units as u
import matplotlib
from scipy.signal import resample
if 'Qt5Agg' in matplotlib.rcsetup.all_backends:
    matplotlib.use('Qt5Agg')

__author__ = 'jundurraga-ucl'


fs = 500.0 * u.Hz
f_s = 25.0 * u.Hz
n_o = fs.value * 11
t = np.arange(0, n_o) / fs
y1 = np.sin(2 * np.pi * u.rad * f_s * t + np.pi * u.rad / 2) * u.uV
y2 = np.sin(2 * np.pi * u.rad * 5 * f_s * t) * u.uV
y = np.tile(np.array(y1 + y2) + 0e6 + 0 * np.random.random(t.shape), (4, 1)).T * u.uV
new_fs = 250 * u.Hz
y[3 * y.shape[0]//4::, :] = 0

yf = np.fft.fft(y, axis=0)
freq = np.arange(len(yf)) * fs / len(yf)
yc, factor = eeg_resampling(x=y.copy(), new_fs=new_fs, fs=fs, blocks=8)

new_fs = fs * factor
new_time = np.arange(0, yc.shape[0]) / new_fs
plt.plot(t, y[:, 0], label='original', color='b')
plt.plot(t, y[:, 0], 'bo')
plt.plot(t, y1, label='h1')
plt.plot(t, y2, label='h2')
plt.plot(np.squeeze(new_time), yc[:, 0], 'r', label='Resampled_1')
plt.plot(np.squeeze(new_time), yc[:, 0], 'rv', label=None)

resampled_scipy = resample(y, new_time.size, t=None, axis=0)
plt.plot(np.squeeze(new_time), resampled_scipy[:, 0], 'g', label='resampled-scipy')
plt.plot(np.squeeze(new_time), resampled_scipy[:, 0], 'gv', label=None)
plt.legend()
plt.show()
plt.show()

freq_1 = np.arange(0, y.shape[0]) * fs / y.shape[0]
yfft1 = 2 * np.abs(np.fft.fft(y, axis=0)) / y.shape[0]
freq_2 = np.arange(0, yc.shape[0]) * new_fs / yc.shape[0]
yfft2 = 2 * np.abs(np.fft.fft(yc, axis=0)) / yc.shape[0]
freq_4 = np.arange(0, resampled_scipy.shape[0]) * new_fs / resampled_scipy.shape[0]
yfft4 = 2 * np.abs(np.fft.fft(resampled_scipy, axis=0)) / resampled_scipy.shape[0]

plt.figure()
plt.plot(freq_1, yfft1[:, 0], label='original')
plt.plot(freq_2, yfft2[:, 0], label='original-m1')
plt.plot(freq_4, yfft4[:, 0], label='scipy')
plt.legend()
plt.xlim([0, new_fs.value / 2])
plt.show()



import numpy as np
from peegy.processing.tools.filters import eegFiltering as eegf
import scipy.signal as signal
import time
import matplotlib.pyplot as plt
import peegy.processing.tools.multiprocessing.multiprocessesing_filter as mp
import astropy.units as u
# Enable below for interactive backend
import matplotlib
if 'Qt5Agg' in matplotlib.rcsetup.all_backends:
    matplotlib.use('Qt5Agg')

fs = 1000.0 * u.Hz
f1 = 10.0 * u.Hz
f2 = 43.0 * u.Hz
low_pass = 16.115 * u.Hz
high_pass = 3.885 * u.Hz
low = 2 * low_pass / fs
high = 2 * high_pass / fs
t = np.arange(0, 70000) / fs
s = np.sin(2 * np.pi * u.rad * f1 * t) + np.sin(2 * np.pi * u.rad * f2 * t)
data = np.tile(s, [10, 1]).T

_b_l_1 = eegf.bandpass_fir_win(fs=fs, high_pass=high_pass, low_pass=low_pass, transition_width=1 * u.Hz,
                               filt_filt_cutoff=False)
n_passes = 2
_b_l_2 = eegf.bandpass_fir_win(fs=fs, high_pass=high_pass, low_pass=low_pass, transition_width=1 * u.Hz,
                               filt_filt_cutoff=True, n_passes=n_passes)
w, h_1 = signal.freqz(_b_l_1, worN=36000)
w, h_2 = signal.freqz(_b_l_2, worN=36000)
plt.plot((w/np.pi)*fs*0.5, 20 * np.log10(np.abs(h_2)), linewidth=2)
plt.plot((w/np.pi)*fs*0.5, 20 * np.log10(np.abs(h_2) ** n_passes), linewidth=2)
plt.xlabel('Frequency (Hz)')
plt.show()

_data = data.copy()
time1 = time.time()
filtered_data_1 = mp.filt_data(_data, _b_l_2)
print(('multiprocessing filter took %f [ms]' % ((time.time()-time1) * 1000.0)))


_data = data.copy()
time1 = time.time()
filtered_data_2 = eegf.filt_worker_ovs(data=_data, pos_ini=0, pos_end=-1, b=_b_l_2)
print(('Overlap save filter took %f [ms]' % ((time.time()-time1) * 1000.0)))

_data = data.copy()
time1 = time.time()
filtered_data_3 = mp.filt_filt_data(data=_data, b=_b_l_2)
print(('Filtfilt filter took %f [ms]' % ((time.time()-time1) * 1000.0)))

_data = data.copy()
time1 = time.time()
filtered_data_4 = mp.filt_data(data=_data, b=_b_l_2, onset_padding=False)
print(('Filtfilt filter took %f [ms]' % ((time.time()-time1) * 1000.0)))


_data = data.copy()
time1 = time.time()
filtered_signal_standard = signal.lfilter(_b_l_2, [1], _data, axis=0)
print(('Standard filter took %f [ms]' % ((time.time()-time1) * 1000.0)))

plt.plot(t, data[:, 0], label='unfiltered')
plt.plot(t, filtered_data_1[:, 0], label='parallel filt ovs')
plt.plot(t, filtered_data_2[:, 0], label='filt ovs')
plt.plot(t, filtered_data_3[:, 0], label='filtfilt')
plt.plot(t, filtered_data_4[:, 0], label='filt no padding')
plt.plot(t, filtered_signal_standard[:, 0], label='scipy filt')
plt.legend()
plt.show()



# -*- coding: utf-8 -*-
"""
Test various methods to denoise data
"""
# import initExample  ## Add path to library (just for examples; you do not need this)

from peegy.processing.tools.eeg_epoch_operators import et_multi_shift, et_covariance, et_time_shifted_xcovariance
import numpy as np

a = np.arange(0, 100.0).reshape(10, -1, 5, order='F')
b = et_multi_shift(a, np.array([0]))
print(b)

b = et_covariance(a, shifts=np.array([2]), w=np.ones(a.shape) * 2.0)
print(b)

c = et_time_shifted_xcovariance(a, a - 1, shifts=np.array([2]), w=np.ones(a.shape) * 2.0)
print(c)

c = et_time_shifted_xcovariance(a, a - 1, shifts=np.array([2, 3]), w=np.ones(a.shape) * 2.0)
print(c)



import numpy
import numpy as np
import multiprocessing
import pyfftw
import astropy.units as u
from tqdm import tqdm
from .. multiprocessing.multiprocessesing_filter import filt_data
from .. filters.eegFiltering import bandpass_fir_win
__author__ = 'jundurraga'


def eeg_resampling(x: np.array = np.array([]),
                   new_fs: u.Quantity = None,
                   fs: u.Quantity = None,
                   blocks=8,
                   padding: bool = True,
                   padding_width: int = None) -> np.array:
    factor = new_fs / fs
    if factor == 1.0:
        return x, factor
    n_samples = x.shape[0]
    if padding:
        if factor < 1.0:
            _b = bandpass_fir_win(high_pass=None,
                                  low_pass=new_fs / 2 * (1 - 0.1),
                                  fs=fs,
                                  transition_width=new_fs / 2 * 0.1)
            _b = _b * u.dimensionless_unscaled
            # apply filter to avoid edge effects
            x = filt_data(data=x, b=_b)
        # padding is using to reduce ringing when down sampling in the frequency domain.
        # ensure that padding length is that so that the input fft size is a multiple of the output fft. This will
        # ensure that down-sampling multiples of fs will as desired
        if padding_width is None:
            padding_width = x.shape[0] // 4
        if factor < 1.0:
            step = np.ceil(1 / factor).astype(int)
        else:
            step = np.floor(factor).astype(int)

        nfft_points = int(np.ceil((x.shape[0] + padding_width) / step) * step)
        pad_width = nfft_points - x.shape[0]
        a1 = x[-1, :]
        a2 = x[0, :]
        padded_samples = (a2 - a1) * np.arange(pad_width)[:, None] / pad_width + a1
        x = np.vstack((x, padded_samples))
    blocks = np.minimum(x.shape[1], blocks)
    updated_nfft = int(x.shape[0] / fs * new_fs)
    xfilt = np.zeros((updated_nfft, x.shape[1]))
    for b in tqdm(np.arange((np.ceil(x.shape[1]) / blocks)), desc='Resampling'):
        pos_ini = int(b * blocks)
        pos_end = int(np.minimum((b + 1) * blocks, x.shape[1]) + 1)
        fft = pyfftw.builders.rfft(x[:, pos_ini:pos_end],
                                   overwrite_input=False,
                                   planner_effort='FFTW_ESTIMATE',
                                   axis=0,
                                   threads=multiprocessing.cpu_count())
        fft_data = fft()
        _fft_shifted = fft_data
        ifft = pyfftw.builders.irfft(_fft_shifted,
                                     n=updated_nfft,
                                     overwrite_input=False,
                                     planner_effort='FFTW_ESTIMATE',
                                     axis=0,
                                     threads=multiprocessing.cpu_count())
        sub_set = ifft()
        xfilt[:, pos_ini:pos_end] = sub_set
        if pos_end == x.shape[1]:
            break
    _scaling_factor = xfilt.shape[0] / x.shape[0]
    _ini = 0
    _end = int(n_samples * _scaling_factor)
    xfilt = xfilt[_ini: _end, :]
    return xfilt * _scaling_factor * x.unit, _scaling_factor



"""
.. _tut-acc-standard-vs-weighted-average-sim:

################################################
ACC standard vs. weighted average (Simulated)
################################################
This example simulates an ACC response and compares the use of standard with weighted average.

"""
# Enable below for interactive backend
# import matplotlib
# if 'Qt5Agg' in matplotlib.rcsetup.all_backends:
#    matplotlib.use('Qt5Agg')
from peegy.processing.pipe.pipeline import PipePool
from peegy.processing.tools.detection.definitions import TimePeakWindow, PeakToPeakMeasure, TimeROI
from peegy.processing.pipe.detection import PeakDetectionTimeDomain
from peegy.processing.pipe.general import ReferenceData, FilterData, RegressOutEOG, ReSampling, AutoRemoveBadChannels
from peegy.processing.pipe.epochs import AverageEpochs, EpochData, SortEpochs
from peegy.processing.pipe.plot import PlotWaveforms, PlotTopographicMap
from peegy.processing.pipe.spatial_filtering import CreateAndApplySpatialFilter
from peegy.processing.pipe.statistics import HotellingT2Test
from peegy.processing.pipe.simulate import GenerateInputData
from peegy.processing.tools.template_generator.auditory_waveforms import aep
from peegy.processing.pipe.storage import MeasurementInformation, SubjectInformation, SaveToDatabase
import os
import astropy.units as u
import numpy as np

# %%
# Define analysis windows for peak detection
# -------------------------------------------
# First we define the time windows that would be used to detect time-domain peaks.


tw = np.array([TimePeakWindow(ini_time=50 * u.ms, end_ref='N1', label='P1', positive_peak=True),
               TimePeakWindow(ini_time=100 * u.ms, end_time=200 * u.ms, label='N1', positive_peak=False),
               TimePeakWindow(ini_ref='N1', end_time=300 * u.ms, label='P2', positive_peak=True)]
              )

# %%
# Define peak-to-peak and region-of-interest measures
# --------------------------------------------------------
# Next, we define the peak-to-peak measures we want to save. These are based on the peaks that would be searched within
# the time windows defined in the previous step.
# TimeROI defines generic time windows that would be used to compute different statistics requiring a specific window,
# for example, Hotelling T2 tests in the time-domain.


pm = np.array([PeakToPeakMeasure(ini_peak='N1', end_peak='P2')])
roi_windows = np.array([TimeROI(ini_time=100.0 * u.ms, end_time=250.0 * u.ms, label="itd_snr"),
                        TimeROI(ini_time=0.0 * u.ms, end_time=400.0 * u.ms, label="itd_snr")])


# %%
# Generating some data
# ------------------------
# Now we generate a target signal, in this example, and auditory N1-P2 (ACC) signal.

n_channels=32
fs = 512.0 * u.Hz
template_waveform, _ = aep(fs=fs)
event_times = np.arange(0, 100.0, 1.0) * u.s
reader = GenerateInputData(template_waveform=template_waveform,
                           fs=fs,
                           n_channels=n_channels,
                           mixing_matrix=np.diag(np.arange(n_channels))/n_channels,
                           layout_file_name='biosemi32.lay',
                           snr=0.05,
                           include_eog_events=True,
                           include_non_stationary_noise_events=False,
                           noise_events_interval=9.0 * u.s,
                           noise_events_duration=3.0 * u.s,
                           noise_events_power_delta_db=25.0,
                           noise_seed=0,
                           event_times=event_times,
                           event_code=1.0,
                           figures_subset_folder='acc_standard_vs_weighted_test')
reader.run()

# %%
# Start the pipeline
# =======================
# Now we proceed with our basic processing pipeline
pipeline = PipePool()
pipeline['referenced'] = ReferenceData(reader, reference_channels=['Cz'],
                                       invert_polarity=True)
pipeline['channel_cleaned'] = AutoRemoveBadChannels(pipeline['referenced'])
pipeline['down_sampled'] = ReSampling(pipeline['channel_cleaned'],
                                      new_sampling_rate=256. * u.Hz)
pipeline['eog_removed'] = RegressOutEOG(pipeline['down_sampled'],
                                        ref_channel_labels=['EOG1'])

pipeline['time_filtered_data'] = FilterData(pipeline['eog_removed'],
                                            high_pass=2.0 * u.Hz,
                                            low_pass=30.0 * u.Hz)
pipeline.run()

# %%
# Get Epochs
# -------------
# We partition the data into epochs or trials based on the event code used.


pipeline['time_epochs'] = EpochData(pipeline['time_filtered_data'],
                                    event_code=1.0)
pipeline['dss_time_epochs'] = CreateAndApplySpatialFilter(pipeline['time_epochs'],
                                                          weight_data=True,
                                                          weight_across_epochs=False,
                                                          block_size=1)

# %%
# Compute HT2 statistics on weighted data
# ========================================
# Using the raw epochs we estimate Hotelling-T2 statistics using weighted averaging


pipeline['ht2_weighted_average'] = HotellingT2Test(pipeline['dss_time_epochs'],
                                                   roi_windows=roi_windows,
                                                   weight_data=True)

# %%
# Sort cleaned epochs
# ----------------------
# Data is now sorted base on their standard deviation (or RMS)

pipeline['sorted_dss_epochs'] = SortEpochs(pipeline['dss_time_epochs'])
pipeline.run()

# %%
# Compute average responses
# ----------------------------
# We compute weighted average on sorted epochs


pipeline['weighted_time_average'] = AverageEpochs(pipeline['sorted_dss_epochs'],
                                                  weight_across_epochs=False,
                                                  weighted_average=True,
                                                  roi_windows=roi_windows)

pipeline.run()

# %%
# Detect peaks
# ---------------
# Using the average data, we proceed detecting peaks


pipeline['peak_detection_weighted_ave'] = PeakDetectionTimeDomain(pipeline['weighted_time_average'],
                                                                  time_peak_windows=tw,
                                                                  peak_to_peak_measures=pm)

# %%
# Show some waveforms
# ----------------------


pipeline['topographic_map_1'] = PlotTopographicMap(pipeline['peak_detection_weighted_ave'],
                                                   topographic_channels=np.array(['C4', 'CP2']),
                                                   plot_x_lim=[0, 0.8],
                                                   plot_y_lim=[-3, 3],
                                                   user_naming_rule='_weighted_time_average',
                                                   return_figures=True)
pipeline.run()


# %%
# Compute HT2 statistics on standard unweighted data
# ==================================================
# Using the raw epochs we estimate Hotelling-T2 statistics using standard averaging


pipeline['ht2_standard_average'] = HotellingT2Test(pipeline['dss_time_epochs'],
                                                   roi_windows=roi_windows,
                                                   weight_data=True)
pipeline.run()

# %%
# Compute average responses
# -------------------------
# We compute standard average

pipeline['standard_time_average'] = AverageEpochs(pipeline['time_epochs'],
                                                  roi_windows=roi_windows,
                                                  weighted_average=False)

pipeline.run()


# %%
# Detect peaks
# ------------
# Using the average data, we proceed detecting peaks


pipeline['peak_detection_standard_ave'] = PeakDetectionTimeDomain(pipeline['standard_time_average'],
                                                                  time_peak_windows=tw,
                                                                  peak_to_peak_measures=pm)

# %%
# Show some waveforms
# -------------------

pipeline['topographic_map_2'] = PlotTopographicMap(pipeline['peak_detection_standard_ave'],
                                                   topographic_channels=np.array(['C4', 'CP2']),
                                                   plot_x_lim=[0, 0.8],
                                                   plot_y_lim=[-3, 3],
                                                   return_figures=True,
                                                   user_naming_rule='_standard_time_average')

pipeline['plotter'] = PlotWaveforms(pipeline['peak_detection_standard_ave'],
                                    overlay=[pipeline['peak_detection_weighted_ave']],
                                    plot_x_lim=[0, 0.8],
                                    return_figures=True,
                                    user_naming_rule='_standard_vs_w_average')

pipeline.run()

# %%
# Get generated data and save to database
# =======================================
# Now we save our data to a database

subject_info = SubjectInformation(subject_id='Test_Subject')
measurement_info = MeasurementInformation(
    date='Today',
    experiment='sim')

_parameters = {'Type': 'ACC'}
database_path = reader.input_node.paths.file_directory + os.sep + 'acc_standard_vs_weighted_test_data.sqlite'
pipeline['database'] = SaveToDatabase(database_path=database_path,
                                      measurement_information=measurement_info,
                                      subject_information=subject_info,
                                      recording_information={'recording_device': 'dummy_device'},
                                      stimuli_information=_parameters,
                                      processes_list=[pipeline['peak_detection_standard_ave'],
                                                      pipeline['peak_detection_weighted_ave'],
                                                      pipeline['ht2_standard_average'],
                                                      pipeline['ht2_weighted_average']])
pipeline.run()

# %%
# Generate pipeline diagram
# ------------------------------------
pipeline.diagram(file_name=reader.output_node.paths.figures_current_dir + 'pipeline.png',
                 return_figure=True,
                 dpi=600)



"""
.. _tut-acc-video-pipeline-sim:

####################################################################
ACC Standard analysis with video (Simulated)
####################################################################

In this example we simulate an ACC response and detect the peaks using a standard pipeline.
We generate a video showing the topographic map over time.

"""
# Enable below for interactive backend
# import matplotlib
# if 'Qt5Agg' in matplotlib.rcsetup.all_backends:
#    matplotlib.use('Qt5Agg')
from peegy.processing.pipe.pipeline import PipePool
from peegy.processing.pipe.detection import PeakDetectionTimeDomain
from peegy.processing.pipe.general import ReferenceData, FilterData, RegressOutEOG, ReSampling, AutoRemoveBadChannels, \
    BaselineCorrection
from peegy.processing.pipe.epochs import AverageEpochs, EpochData
from peegy.processing.pipe.attach import AppendGFPChannel
from peegy.processing.pipe.plot import PlotWaveforms, PlotTopographicMap, TopographicMapVideo
from peegy.processing.pipe.spatial_filtering import CreateAndApplySpatialFilter
from peegy.processing.pipe.statistics import HotellingT2Test
from peegy.processing.pipe.simulate import GenerateInputData
from peegy.processing.tools.detection.definitions import TimePeakWindow, PeakToPeakMeasure, TimeROI
from peegy.processing.tools.template_generator.auditory_waveforms import aep
from peegy.processing.pipe.storage import MeasurementInformation, SubjectInformation, SaveToDatabase
import astropy.units as u
import numpy as np
import os

# %%
# Time windows
# ============================
# Time windows are used to define where to look for different peaks


tw = np.array([TimePeakWindow(ini_time=50 * u.ms, end_ref='N1', label='P1', positive_peak=True,
                              exclude_channels=['GFP']),
               TimePeakWindow(ini_time=80 * u.ms, end_time=200 * u.ms, label='N1', positive_peak=False,
                              exclude_channels=['GFP']),
               TimePeakWindow(ini_ref='N1', end_time=300 * u.ms, label='P2', positive_peak=True,
                              exclude_channels=['GFP']),
               TimePeakWindow(ini_time=50 * u.ms, end_time=150 * u.ms, label='gfp1', positive_peak=True,
                              target_channels=['GFP']),
               TimePeakWindow(ini_time=150 * u.ms, end_time=500 * u.ms, label='gfp2', positive_peak=True,
                              target_channels=['GFP'])])
# %%
# Peak-to-peak Measures
# ============================
# Peak-to-peak measures are defined based on the labels of the TimePeaks defined above.


pm = np.array([PeakToPeakMeasure(ini_peak='N1', end_peak='P2')])

# %%
# Time regions of interest
# ============================
# TimeROI are defined as time regions where different measures will be performed, e.g. SNR measure or statistical
# measures


roi_windows = np.array([TimeROI(ini_time=100.0 * u.ms, end_time=250.0 * u.ms, label="acc_snr")])

# %%
# Generate some data
# ====================
# First we generate some ACC data

n_channels=32
fs = 512.0 * u.Hz
template_waveform, _ = aep(fs=fs)
event_times = np.arange(0, 100.0, 1.0) * u.s
reader = GenerateInputData(template_waveform=template_waveform,
                           fs=fs,
                           n_channels=n_channels,
                           mixing_matrix=np.diag(np.arange(n_channels))/n_channels,
                           layout_file_name='biosemi32.lay',
                           snr=0.1,
                           event_times=event_times,
                           event_code=1.0,
                           figures_subset_folder='acc_test_video',
                           include_eog_events=True,
                           noise_seed=0
                           )
reader.run()

# %%
# Start the pipeline
# ====================
# Now we proceed with our basic processing pipeline


pipeline = PipePool()
pipeline['referenced'] = ReferenceData(reader,
                                       reference_channels=['Cz'],
                                       invert_polarity=True)

pipeline['channel_cleaned'] = AutoRemoveBadChannels(pipeline['referenced'])

pipeline['down_sampled'] = ReSampling(pipeline['channel_cleaned'],
                                      new_sampling_rate=256. * u.Hz)
pipeline['eog_removed'] = RegressOutEOG(pipeline['down_sampled'],
                                        ref_channel_labels=['EOG1'],
                                        method='template',
                                        n_iterations=10,
                                        use_initial_template=True)
pipeline.run()
# %%
# Show EOG removal Output
# --------------------------------------

pipeline['eog_removed'].plot(plot_input=True,
                             plot_output=True,
                             ch_to_plot=['CP1', 'CP5', 'P7'],
                             interactive=False)
# %%
# Continue with the pipeline
# --------------------------------------

pipeline['time_filtered_data'] = FilterData(pipeline['eog_removed'],
                                            high_pass=2.0 * u.Hz,
                                            low_pass=30.0 * u.Hz)

pipeline.run()

# %%
# Get Epochs
# --------------------------------------
# We partition the data into epochs or trials based on the event code used.


pipeline['time_epochs'] = EpochData(pipeline['time_filtered_data'],
                                    event_code=1.0)
pipeline.run()

# %%
# Get DSS components for EOG free
# --------------------------------------
# Compute spatial filter based on EOG free epochs

pipeline['time_epochs_dss'] = CreateAndApplySpatialFilter(pipeline['time_epochs'],
                                                          return_figures=True)

# %%
# Compute global field power (GFP)
# --------------------------------------
# We compute the GFP across channels and epochs. A new channel with the GFP is appended to the data


pipeline['time_epochs_with_gfp'] = AppendGFPChannel(pipeline['time_epochs'])
pipeline['time_epochs_with_gfp_dss'] = AppendGFPChannel(pipeline['time_epochs_dss'])
pipeline.run()

# %%
# Base line correction
# --------------------------------------
# We correct the initial segment by removing the mean in the first 20 ms

pipeline['time_epochs_with_baseline_corrected'] = BaselineCorrection(pipeline['time_epochs_with_gfp'],
                                                                     ini_time=0 * u.ms,
                                                                     end_time=20 * u.ms)
pipeline['time_epochs_with_baseline_corrected_dss'] = BaselineCorrection(pipeline['time_epochs_with_gfp_dss'],
                                                                         ini_time=0 * u.ms,
                                                                         end_time=20 * u.ms)
pipeline.run()

# %%
# Compute HT2 statistics
# --------------------------------------
# Using the raw epochs we estimate Hotelling-T2 statistics


pipeline['ht2'] = HotellingT2Test(pipeline['time_epochs_with_baseline_corrected'],
                                  roi_windows=roi_windows)
pipeline['ht2_dss'] = HotellingT2Test(pipeline['time_epochs_with_baseline_corrected_dss'],
                                      roi_windows=roi_windows)
pipeline.run()

# %%
# Compute average responses
# --------------------------------------
# We compute weighted average on epochs with and without spatial filtering (DSS)


pipeline['time_average_normal'] = AverageEpochs(pipeline['time_epochs_with_baseline_corrected'],
                                                roi_windows=roi_windows,
                                                weighted_average=True)
pipeline['time_average_normal_dss'] = AverageEpochs(pipeline['time_epochs_with_baseline_corrected_dss'],
                                                    roi_windows=roi_windows,
                                                    weighted_average=True)
pipeline.run()

# %%
# Detect peaks
# --------------------------------------
# Using the average data, we proceed detecting peaks


pipeline['data_with_peaks'] = PeakDetectionTimeDomain(pipeline['time_average_normal'],
                                                      time_peak_windows=tw,
                                                      peak_to_peak_measures=pm)
pipeline['data_with_peaks_dss'] = PeakDetectionTimeDomain(pipeline['time_average_normal_dss'],
                                                          time_peak_windows=tw,
                                                          peak_to_peak_measures=pm)

# %%
# Show some waveforms
# --------------------------------------


pipeline['plotter'] = PlotWaveforms(pipeline['data_with_peaks'],
                                    ch_to_plot=np.array(['C4', 'CP2', 'GFP']),
                                    overlay=[pipeline['data_with_peaks_dss']],
                                    statistical_test='f_test_time',
                                    show_following_stats=['f', 'rn'],
                                    return_figures=True,
                                    )
pipeline.run()

# %%
# Show some topographic maps
# --------------------------------------


pipeline['topographic_map_1'] = PlotTopographicMap(pipeline['data_with_peaks'],
                                                   topographic_channels=['T8'],
                                                   plot_x_lim=[0, 0.8],
                                                   plot_y_lim=[-3, 3],
                                                   user_naming_rule='standard',
                                                   return_figures=True)
pipeline['topographic_map_2'] = PlotTopographicMap(pipeline['data_with_peaks_dss'],
                                                   topographic_channels=['T8'],
                                                   plot_x_lim=[0, 0.8],
                                                   plot_y_lim=[-3, 3],
                                                   user_naming_rule='dss',
                                                   return_figures=True)
pipeline.run()

# %%
# Generate a video showing the topographic map over time
# ---------------------------------------------------------


pipeline['video'] = TopographicMapVideo(pipeline['data_with_peaks_dss'],
                                        topographic_channels=['T8', 'GFP'],
                                        step_size=15 * u.ms,
                                        plot_x_lim=[0, 0.8],
                                        plot_y_lim=[-3, 3],
                                        user_naming_rule='dss')
pipeline.run()

# %%
# Get generated data and save to database
# -----------------------------------------
# Now we save our data to a database

subject_info = SubjectInformation(subject_id='Test_Subject')
measurement_info = MeasurementInformation(
    date='Today',
    experiment='sim')

_parameters = {'Type': 'ACC'}
database_path = reader.input_node.paths.file_directory + os.sep + 'acc_test_data.sqlite'
pipeline['database'] = SaveToDatabase(database_path=database_path,
                                      measurement_information=measurement_info,
                                      subject_information=subject_info,
                                      recording_information={'recording_device': 'dummy_device'},
                                      stimuli_information=_parameters,
                                      processes_list=[pipeline['data_with_peaks'],
                                                      pipeline['data_with_peaks_dss'],
                                                      pipeline['ht2'],
                                                      pipeline['ht2_dss']]
                                      )
pipeline.run()

# %%
# Generate pipeline diagram
# ------------------------------------
pipeline.diagram(file_name=reader.output_node.paths.figures_current_dir + 'pipeline.png',
                 return_figure=True,
                 dpi=600)



import numpy as np
from peegy.processing.tools.filters import eegFiltering as eegf
import scipy.signal as signal
import time
import matplotlib.pyplot as plt

fs = 1000.0
f1 = 10.0
f2 = 43.0
low_pass = 50.0
high_pass = 40.0
low = 2 * low_pass / fs
high = 2 * high_pass / fs
t = np.arange(0, 7000000) / fs
s = np.sin(2 * np.pi * f1 * t) + np.sin(2 * np.pi * f2 * t)
data = np.tile(s, [12, 1]).T

# _b_l = eegf.bandpass_fir_win(fs=fs, low_pass=low_pass)
# w, h = signal.freqz(_b_l, worN=8000)
# plt.plot((w/np.pi)*fs*0.5, np.abs(h), linewidth=2)
# plt.xlabel('Frequency (Hz)')

# _b_l = eegf.bandpass_fir_win(fs=fs, high_pass=high_pass)
# w, h = signal.freqz(_b_l, worN=8000)
# plt.plot((w/np.pi)*fs*0.5, np.abs(h), linewidth=2)
# plt.xlabel('Frequency (Hz)')

_b_l = eegf.bandpass_fir_win(fs=fs, high_pass=high_pass, low_pass=low_pass)
w, h = signal.freqz(_b_l, worN=8000)
plt.plot((w/np.pi)*fs*0.5, np.abs(h), linewidth=2)
plt.xlabel('Frequency (Hz)')


_ori_data = data.copy()
filtered_data = eegf.filt_filt_worker_ovs(data=_ori_data.copy(), pos_ini=0, pos_end=None, b=_b_l)
filtered_data2 = eegf.filt_worker_ovs(data=_ori_data.copy(), pos_ini=0, pos_end=None, b=_b_l)
time1 = time.time()
filtered_signal_standard = signal.filtfilt(_b_l, 1, data, axis=0)
time2 = time.time()
print(('standard filter took %f [ms]' % ((time2-time1) * 1000.0)))
plt.figure()
plt.plot(t, _ori_data[:, 0], label='original')
plt.plot(t, filtered_data[:, 0], label='ovs_filt_filt')
plt.plot(t, filtered_signal_standard[:, 0], label='scipy_filt_filt')
plt.plot(t, filtered_data2[:, 0], label='ols_filt')
plt.legend()
plt.show()



import psutil as psu
import numpy as np


def enough_memory(data: np.array = None):
    """
    This function returns true, when there is enough memory in the system to allocate an array of same dimensions as
    data
    :param data: array from which memory will be estimated
    :return:
    """
    _virtual_memory = psu.virtual_memory()
    memory_fraction = _virtual_memory.available / data.nbytes
    return memory_fraction > 1



"""
.. _tut-assr-using-own-raw-data-sim:

#######################################################################
ASSR F-Test Example using 'external' data (time x channels) (Simulated)
#######################################################################

This example shows how you can pass any data to the pipeline (e.g. matlab data, or text file data) in the shape of time
x channels alongside with event times used to epoch the data.
In this example we simulate an ASSR (as it was read from an external file), and we investigate the significance of
the response by means of an F-test that compares the variance of the frequency of interest with the variance of
neighbouring frequency bins.

"""
# Enable below for interactive backend
# import matplotlib
# if 'Qt5Agg' in matplotlib.rcsetup.all_backends:
#    matplotlib.use('Qt5Agg')
from peegy.processing.pipe.pipeline import PipePool
from peegy.processing.pipe.definitions import Domain
from peegy.processing.pipe.io import GenericInputData
from peegy.processing.pipe.statistics import FTest
from peegy.processing.pipe.general import FilterData
from peegy.processing.pipe.epochs import AverageEpochs, EpochData
from peegy.processing.pipe.plot import PlotWaveforms, PlotTopographicMap
from peegy.processing.pipe.spatial_filtering import CreateAndApplySpatialFilter
from peegy.processing.pipe.storage import MeasurementInformation, SubjectInformation, SaveToDatabase
import os
import astropy.units as u
import numpy as np

# %%
# Generate some data (time x channels) and add event times
# ==========================================================
# We generate some data in this example, for the sake of exemplify it. The data is put into GenericInputData class and
# then used in a standard pipeline


fs = 256.0 * u.Hz
epoch_length = 4.0 * u.s
epoch_length = np.ceil(epoch_length * fs) / fs  # fit to fs rate
assr_frequency = 41.0 * u.Hz
assr_frequency = np.ceil(epoch_length * assr_frequency) / epoch_length  # fit to epoch length
test_frequencies = np.array([assr_frequency.value, 2 * assr_frequency.value]) * u.Hz
# here we generate and pass our own data
n_epochs = 20
time = np.arange(0, epoch_length.to(u.s).value * n_epochs, 1 / fs.to(u.Hz).value) * u.s
my_sin = np.sin(2 * np.pi * u.rad * assr_frequency * time)
template_waveform = np.array([my_sin, 0.5 * my_sin, 0.25 * my_sin]).T
# add some noise
np.random.seed(1)
template_waveform = template_waveform + np.random.randn(*template_waveform.shape)
event_times = np.arange(0, n_epochs) * epoch_length
reader = GenericInputData(data=template_waveform,
                          fs=fs,
                          event_times=event_times,
                          event_code=1.0,
                          figures_subset_folder='my_own_raw_assr_f_test')
reader.run()
# %%
# Start the pipeline
# ==========================================================


pipeline = PipePool()
pipeline['time_filtered_data'] = FilterData(reader,
                                            high_pass=2.0 * u.Hz,
                                            low_pass=100.0 * u.Hz)
pipeline.run()

# %%
# Get Epochs
# -----------------------------------------
# We partition the data into epochs or trials based on the event code used.


pipeline['time_epochs'] = EpochData(pipeline['time_filtered_data'],
                                    event_code=1.0,
                                    base_line_correction=False)
pipeline.run()

# %%
# Get DSS components and clean data
# -----------------------------------------
# Compute spatial filter of time epochs


pipeline['dss_time_epochs'] = CreateAndApplySpatialFilter(pipeline['time_epochs'],
                                                          test_frequencies=test_frequencies,
                                                          components_to_plot=np.arange(0, 5),
                                                          projection_domain=Domain.frequency)
pipeline.run()

# %%
# Compute average responses
# -----------------------------------------
# We compute weighted average on epochs with and without spatial filtering (DSS)

pipeline['time_ave_dss'] = AverageEpochs(pipeline['dss_time_epochs'])
pipeline['time_ave'] = AverageEpochs(pipeline['time_epochs'])
pipeline.run()

# %%
# Run F-Test
# -----------------------------------------
# Now we run an F-test to determine if frequency component is significant


pipeline['fft_ave_f_tests_dss'] = FTest(pipeline['time_ave_dss'],
                                        n_fft=int(epoch_length*fs*1),
                                        test_frequencies=test_frequencies,
                                        delta_frequency=9.0 * u.Hz)
pipeline['fft_ave_f_tests'] = FTest(pipeline['time_ave'],
                                    n_fft=int(epoch_length * fs * 1),
                                    test_frequencies=test_frequencies,
                                    delta_frequency=9. * u.Hz)

# %%
# Show some waveforms
# -----------------------------------------

pipeline['topographic_map_1'] = PlotTopographicMap(pipeline['fft_ave_f_tests_dss'],
                                                   plot_x_lim=[0, 90],
                                                   plot_y_lim=[0, 4],
                                                   topographic_channels=np.array(['CH_0', 'CH_1', 'CH_2']),
                                                   return_figures=True,
                                                   user_naming_rule='dss')

pipeline['topographic_map_2'] = PlotTopographicMap(pipeline['fft_ave_f_tests'],
                                                   plot_x_lim=[0, 90],
                                                   plot_y_lim=[0, 4],
                                                   topographic_channels=np.array(['CH_0', 'CH_1', 'CH_2']),
                                                   return_figures=True,
                                                   user_naming_rule='standard')

pipeline['plotter'] = PlotWaveforms(pipeline['fft_ave_f_tests'],
                                    ch_to_plot=np.array(['CH_0']),
                                    overlay=[pipeline['fft_ave_f_tests_dss']],
                                    plot_x_lim=[0, 90],
                                    show_following_stats=['frequency_tested', 'f'],
                                    return_figures=True,
                                    user_naming_rule='waveforms',
                                    fig_format='.png')
pipeline.run()

# %%
# Get generated data and save to database
# -----------------------------------------
# We get the measurements we are interested in and save them into a database

subject_info = SubjectInformation(subject_id='Test_Subject')
measurement_info = MeasurementInformation(
    date='Today',
    experiment='sim')
_parameters = {'Type': 'ASSR'}
database_path = reader.input_node.paths.file_directory + os.sep + 'assr_own_raw_data_f_test_data.sqlite'
pipeline['database'] = SaveToDatabase(database_path=database_path,
                                      measurement_information=measurement_info,
                                      subject_information=subject_info,
                                      recording_information={'recording_device': 'dummy_device'},
                                      stimuli_information=_parameters,
                                      processes_list=[pipeline['fft_ave_f_tests'],
                                                      pipeline['fft_ave_f_tests_dss']],
                                      include_waveforms=True
                                      )
pipeline.run()

# %%
# Generate pipeline diagram
# -----------------------------------------
pipeline.diagram(file_name=reader.output_node.paths.figures_current_dir + 'pipeline.png',
                 return_figure=True,
                 dpi=600)



import numpy as np
import pandas as pd
from peegy.definitions.channel_definitions import Domain
import sqlite3
from scipy.interpolate import interp1d
import io

__author__ = 'jundurraga-ucl'


class PandasDataPages(dict):
    """This class manages pandas dataframes appended to it.
    When data frames are appended the PandasDataPages specific dataframe can be called by its name
    """

    def __init__(self):
        super(PandasDataPages, self).__init__()

    def __setitem__(self, name, dataframe: pd.DataFrame = None):
        # optional processing here
        assert isinstance(dataframe, pd.DataFrame)
        name = self.ensure_unique_name(label=name)
        super(PandasDataPages, self).__setitem__(name, dataframe)
        # we add the new item as class variable
        setattr(self, name, dataframe)

    def append(self, item: object, name=None):
        if name is None:
            name = type(item).__name__
        self[name] = item

    def ensure_unique_name(self, label: str = None):
        all_names = [_key for _key in self.keys()]
        _label = label
        count = 0
        while _label in all_names:
            _label = label + '_' + str(count)
            count = count + 1
        if count > 0:
            print('PandasDataPages item "{:}" already exists. Renamed to "{:}"'.format(label, _label))
        return _label

    def __getitem__(self, key):
        return super(PandasDataPages, self).__getitem__(key)


def sqlite_tables_to_pandas(database_path: str = None,
                            tables: [str] = None) -> pd.DataFrame:
    """
    This function will return a pandas dataframe containing the desired tables from a pEEGy .sqlite database.
    A pEEGy database will always contain a table 'subjects', 'measurement_info' , 'stimuli', 'recording'.
    Each subject will be linked to measurement by their id. Similarly, each measurement will be linked to each stimulus
    by its id.
    Any other table, for example, recording, waveforms, hotelling_t2_time, hotelling_t2_freq, f_test_time, f_test_freq,
    or other tables created by the user, will be uniquely related to each subject, measurement, recording, and stimuli
    by id_subject, id_measurement, id_recording, and _id_stimuli, respectively.
    This function will provide a user-friendly pandas dataframe by pooling together this information by indexing the
    corresponding ids to their respective values.
    :param database_path: path to the database from which we will read the tables
    :param tables: a list of strings containing the tables want to read. Make sure these tables are present in the
    database
    :return: a pandas dataframe with the respective tables.
    """
    out = PandasDataPages()
    db = sqlite3.connect(database_path)
    cursor = db.cursor()
    cursor.execute("SELECT name FROM sqlite_master WHERE type='table';")
    all_tables = [_tables[0] for _tables in cursor.fetchall()]
    extra_tables = ''
    extra_outputs = ''
    extra_join = ''
    if 'recording' in all_tables:
        extra_tables = extra_tables.join('JOIN recording REC ON REC.id_measurement = MES.id ')
        extra_outputs = extra_outputs.join('{:}.*, '.format('REC'))
        extra_join = 'and  TAB.id_recording == REC.id'

    for _table in tables:
        df = pd.read_sql_query('SELECT SUB.*, '
                               'MES.*, '
                               'STI.*, '
                               '{:}'
                               'TAB.* '
                               'FROM subjects as SUB '
                               'JOIN measurement_info MES ON (MES.id_subject = SUB.id) '
                               'JOIN stimuli STI ON STI.id_measurement = MES.id '
                               '{:}'
                               'JOIN {:} as TAB ON TAB.id_stimuli == STI.id {:}'.format(extra_outputs,
                                                                                        extra_tables,
                                                                                        _table,
                                                                                        extra_join),
                               db)
        out[_table] = df
    return out


def sqlite_waveforms_to_pandas(database_path: str = None,
                               group_factors: [str] = None,
                               user_query: [str] = None,
                               tables: [str] = None,
                               channels: [str] = None,
                               simuli_columns: list[str] = None,
                               ) -> pd.DataFrame:
    """
    This function will return a pandas dataframe containing the waveforms for the specified group_factors.
    The data is assumed to come from a pEEGy .sqlite database.
    A pEEGy database will always contain a table 'subjects', 'measurement_info' , 'stimuli', 'recording'.
    Each subject will be linked to measurement by their id. Similarly, each measurement will be linked to each stimulus
    by its id.
    Any other table, for example, recording, waveforms, hotelling_t2_time, hotelling_t2_freq, f_test_time, f_test_freq,
    or other tables created by the user, will be uniquely related to each subject, measurement, recording, and stimuli
    by id_subject, id_measurement, id_recording, and _id_stimuli, respectively.
    This function will provide a user-friendly pandas dataframe by pooling together this information with the waveforms.
    For each grouping factor, the waveforms will be pooling together in a ndim numpy array.
    If data have not consistent number of samples, then the x axis of the first waveform for a given domain (time or
    frequency) will be used as the reference. All the rest will be interpolated and sampled to that initial x axis.
    In this way a single matrix will be returned with the data for each grouping factor.
    To avoid this last step, you should make sure that all data stored in the database are epoched having the same
    length (fixed pre_stimulus_interval and post_stimulus_interval).
    :param database_path: path to the database from which we will read the tables
    :param group_factors: a list of strings containing the groups for which you want to pool the data. For example,
    if you want to group all the waveforms from a given stimulus parameter in the stimuli table of the database (e.g.
    Amplitude and Frequency; both of which are columns in the table stimuli) you could define the group factors as
    group_factors = ['Amplitude', 'Frequency']. The returned output will then contain rows grouped by each level within
    each factor whilst the waveforms (x and y columns) will contain the data for grouped for each of these levels
    :param user_query: This parameter can be used to include or exclude data based on a logical condition, e.g.
    'subject_id != "S1"'
    :param tables: list of strings indicating the names of other generated tables that will be join to the output
    dataframe. The tables should contain 'subjects', 'measurement_info' , 'stimuli', 'recording' columns so they can
    be joined.
    :param channels: list of string specifying for which channels you want to extract the waveforms. If empty, all
    channels will be returned.
    :param simuli_columns: if not None, then it should indicate which columns to read (it can include an alias for the
     specific columns to read. For example ['channel as ch', 'subject_id as subject'] or ['channel']
    :return: a pandas dataframe with the data grouped by group_factors.
    """
    db = sqlite3.connect(database_path)
    cursor = db.cursor()
    cursor.execute("SELECT name FROM sqlite_master WHERE type='table';")
    all_tables = [_tables[0] for _tables in cursor.fetchall()]

    if channels is not None:
        channels_str = ','.join(['"{:}"'.format(_ch) for _ch in channels])
        channels_str = 'WAVE.channel IN ({:}) and'.format(channels_str)
    extra_tables = ''
    extra_outputs = ''
    if 'recording' in all_tables:
        extra_tables = extra_tables.join('JOIN recording REC ON REC.id_measurement = MES.id ')
        extra_outputs = extra_outputs.join('{:}.*, '.format('REC'))

    if tables is not None:
        extra_tables = extra_tables.join(['JOIN {:} ON {:}.id_stimuli = STI.id '.format(_tab, _tab) for
                                          _tab in tables])
        extra_outputs = extra_outputs.join(['{:}.*, '.format(_tab) for
                                            _tab in tables])
    if simuli_columns is None:
        sim_query = 'STI.*, '
    else:
        sim_query = ''.join(['STI.{:}, '.format(_acro) for _acro in simuli_columns])

    df = pd.read_sql_query(
        'SELECT SUB.*, '
        'MES.*, '
        '{:}'
        '{:}'
        'WAVE.* '
        'FROM subjects as SUB '
        'JOIN measurement_info MES ON MES.id_subject = SUB.id '
        'JOIN stimuli STI ON STI.id_measurement = MES.id '
        '{:}'
        'INNER JOIN waveforms WAVE ON ({:} '
        'WAVE.id_stimuli = STI.id) '.format(sim_query,
                                            extra_outputs,
                                            extra_tables,
                                            channels_str),
        db)
    df = df.loc[:, ~df.columns.duplicated()]
    if user_query is not None:
        df = df.query(user_query)
    out_pd = pd.DataFrame()
    if group_factors is None:
        group_factors = ['domain']
    else:
        group_factors = list(set.union(set(group_factors), set(['domain'])))

    groups = df.groupby(group_factors)
    n_s_time = None
    x_time = None
    n_s_freq = None
    x_freq = None
    for _group_value, _group in groups:
        y = np.array([])
        x = np.array([])
        for index, row in _group.iterrows():
            try:
                x_data = np.load(io.BytesIO(row['x']))
                y_data = np.load(io.BytesIO(row['y']))
            except ValueError:
                x_data = np.frombuffer(io.BytesIO(row['x']).read())
                y_data = np.frombuffer(io.BytesIO(row['y']).read())
            except ValueError:
                print("Could not read data from database")

            x_unit = row['x_unit']
            y_unit = row['y_unit']
            if row['domain'] == Domain.time and n_s_time is None:
                n_s_time = x_data.size
                x_time = x_data
            if row['domain'] == Domain.frequency and n_s_freq is None:
                n_s_freq = x_data.size
                x_freq = x_data
            if row['domain'] == Domain.time and n_s_time != x_data.size:
                f = interp1d(x_data, y_data, fill_value="extrapolate")
                y_data = f(x_time)
                x_data = x_time
            if row['domain'] == Domain.frequency and n_s_freq != x_data.size:
                f = interp1d(x_data, y_data, fill_value="extrapolate")
                y_data = f(x_freq)
                x_data = x_freq
            if y.size == 0:
                y = y_data
                x = x_data
            else:
                y = np.vstack((y, y_data))
            if row['domain'] == Domain.time:
                if 'fs' in row.index:
                    fs = row['fs']
                else:
                    fs = 1 / np.mean(np.diff(x))
            if row['domain'] == Domain.frequency:
                # the fs here is a guess assuming the rfft has all samples
                n = y_data.shape[0]
                if np.mod(n, 2) == 1:
                    time_size = 2 * (n - 1)
                else:
                    time_size = 2 * n + 1
                if 'fs' in row.index:
                    fs = row['fs']
                else:
                    fs = time_size / (1 / np.mean(np.diff(x)))
        if len(group_factors) == 1:
            pars = dict(list(zip(group_factors, [_group_value])))
        else:
            pars = dict(list(zip(group_factors, _group_value)))
        out_pd = pd.concat([out_pd, pd.DataFrame([dict(pars, **{'x': x,
                                                                'y': y.T,
                                                                'x_fs': fs,
                                                                'x_unit': x_unit,
                                                                'y_unit': y_unit
                                                                })])],
                           ignore_index=True)
    return out_pd



import numpy as np
import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle
from peegy.processing.tools.detection.time_domain_tools import get_channel_peaks_and_windows
from peegy.definitions.eeg_definitions import EegChannel
from peegy.definitions.channel_definitions import Domain
import os
from peegy.plot.eeg_plot_tools import interpolate_potential_fields
import gc
import matplotlib.gridspec as gridspec
from operator import itemgetter
from peegy.processing.pipe.definitions import DataNode
import astropy.units as u
from matplotlib.pyplot import cm
from astropy.visualization import quantity_support   # , time_support
from peegy.tools.units.unit_tools import set_default_unit
import warnings
warnings.filterwarnings("ignore", message="No contour levels were found within the data range.")
quantity_support()
# time_support()


def get_potential_fields(epochs_ave: DataNode = DataNode(),
                         x_val: np.array = np.array([]),
                         domain: Domain = Domain.time,
                         **kwargs):
    x_val = np.atleast_1d(x_val)
    if domain == Domain.time:
        potentials = np.squeeze(epochs_ave.data[epochs_ave.x_to_samples(x_val), :])
    if domain == Domain.frequency:
        potentials = np.squeeze(np.abs(epochs_ave.data)[epochs_ave.x_to_samples(x_val), :])

    x = np.zeros((potentials.size, 1))
    y = np.zeros((potentials.size, 1))
    z = np.zeros((potentials.size, 1))
    for i, _ch in enumerate(epochs_ave.layout):
        if (_ch.x is not None and not np.isnan(_ch.x)) and (_ch.y is not None and not np.isnan(_ch.y)):
            x[i] = _ch.x
            y[i] = _ch.y
            z[i] = potentials[i]
    interpolated_potentials = None
    max_distance = None
    if x.any() and y.any() and z.any():
        interpolated_potentials, max_distance = interpolate_potential_fields(x, y, z, **kwargs)
    return interpolated_potentials, max_distance


def plot_single_channels(ave_data: [DataNode] = None,
                         channels: np.array([str]) = np.array([]),
                         figure_dir_path: str = None,
                         figure_basename: str = None,
                         fig_format: str = '.png',
                         statistical_test: str = None,
                         show_following_stats: [str] = None,
                         offset_step: u.quantity.Quantity = None,
                         save_to_file: bool = True,
                         title: str = '',
                         x_lim: [float, float] = None,
                         y_lim: [float, float] = None,
                         show_peaks: bool = True,
                         show_labels: bool = True,
                         fontsize=8):
    """
    :param ave_data: list of DataNode class containing data to plot
    :param figure_dir_path: directory path where figure will be saved
    :param figure_basename: output file name
    :param fig_format: output format of generated figure (.pdf, .png)
    :param statistical_test: name of the statistical test used to extract data
    :param show_following_stats: list of string indicating which parameters found in statistical_tests are shown
    :param channels: index or labels of channels to be plotted
    :param offset_step: constant to vertically separate each channel
    :param save_to_file: if True, figure will be saved into figure_dir_path
    :param title: desired figure title
    :param x_lim: plot x limits
    :param y_lim: plot x limits
    :param show_peaks: if True, detected peaks will be shown
    :param show_labels: if True, peak labels will be shown
    :param fontsize: fontsize of text in figures
    :return:
    """
    # plot individual channels
    font = {'size': 6}
    fig = plt.figure()
    inch = 2.54
    fig.set_size_inches(12 / inch, 8 / inch)
    plt.rc('font', **font)
    colors = cm.tab10.colors[0: len(ave_data)]
    # get list with all channels labels
    all_labels = np.array([_layout.label for _data in ave_data for _layout in _data.layout])
    _, _idx = np.unique(all_labels, return_index=True)
    all_labels = all_labels[np.sort(_idx)]
    if channels is not None:
        if isinstance(channels, list):
            channels = np.array(channels)
        if isinstance(channels[0], str):
            all_labels = np.intersect1d(all_labels, channels)
            # ensure unique labels and remove repeated in same order
            _, idx_channels = np.unique(channels, return_index=True)
            channels = channels[np.sort(idx_channels)]
            idx = []
            for _label in channels:
                _idx = np.argwhere(_label == all_labels)
                if _idx.size:
                    idx.append(_idx.squeeze())
            idx = np.array(idx)
            all_labels = all_labels[idx]
        else:
            _aviable_channels = np.minimum(len(all_labels) - 1, channels)
            _, idx_channels = np.unique(_aviable_channels, return_index=True)
            _aviable_channels = _aviable_channels[np.sort(idx_channels)]
            all_labels = all_labels[_aviable_channels]
    n_ch_to_plot = len(all_labels)

    for _idx_data, (_ave_data, _color) in enumerate(zip(ave_data, colors)):
        my_stats_text = []
        my_color_text = []
        my_offset_text = []
        if channels is not None and channels.size and isinstance(channels[0], str):
            ch_idx = _ave_data.get_channel_idx_by_label(channels)
        else:
            ch_idx = channels if channels is not None else list(range(_ave_data.data.shape[1]))
            ch_idx = np.minimum(ch_idx, _ave_data.data.shape[1] - 1)
        _, idx = np.unique(ch_idx, return_index=True)
        ch_idx = ch_idx[np.sort(idx)]
        _channels = np.array([ch.label for ch in _ave_data.layout])[ch_idx]

        # initialize offset step based on first input and axis for polar plots
        if _idx_data == 0:
            if _ave_data.domain == Domain.frequency:
                gs = gridspec.GridSpec(n_ch_to_plot, 5)
                ax1 = plt.subplot(gs[::, 0:4])
                ax_in = [None] * n_ch_to_plot
                for _idx in range(n_ch_to_plot):
                    ax_in[_idx] = plt.subplot(gs[_idx, 4], projection='polar')
                    ax_in[_idx].tick_params(axis='x', which='both', bottom=False,
                                            top=False, labelbottom=False)
                    ax_in[_idx].tick_params(axis='y', which='both', right=False,
                                            left=False, labelleft=False)
                    # for pos in ['right', 'top', 'bottom', 'left']:
                    #     ax_in[_idx].spines[pos].set_visible(False)

            else:
                gs = gridspec.GridSpec(1, 1)
                ax1 = plt.subplot(gs[0])

            if offset_step is None:
                offset_step = np.max(np.max(np.abs(_ave_data.data[:, ch_idx]), axis=0))
        offset_vector = np.arange(n_ch_to_plot) * offset_step
        _offset_idx = np.argwhere(np.in1d(all_labels, _channels)).squeeze()
        all_markers = _ave_data.markers
        rn_up, rn_down = None, None

        all_peaks = None
        if _ave_data.peaks is not None:
            all_peaks = _ave_data.peaks
        if _ave_data.domain == Domain.time:
            ave_w = _ave_data.data[:, ch_idx] - offset_vector[_offset_idx].reshape(1, -1)
            x_label = 'Time [{:}]'.format(_ave_data.x.unit)
            rn_up = _ave_data.rn[ch_idx] - offset_vector[_offset_idx].reshape(1, -1)
            rn_down = -_ave_data.rn[ch_idx] - offset_vector[_offset_idx].reshape(1, -1)
        if _ave_data.domain == Domain.frequency:
            ave_w = np.abs(_ave_data.data[:, ch_idx]) - offset_vector[_offset_idx].reshape(1, -1)
            x_label = 'Frequency [{:}]'.format(_ave_data.x.unit)
        if all_peaks is not None and all_peaks.shape[0]:
            _unique_peaks = all_peaks.peak_label.unique()
            peak_color = list(iter(cm.Set1(np.linspace(0, 1, _unique_peaks.size * (1 + len(ave_data))))))
        ax1.plot(_ave_data.x, ave_w, linewidth=0.5, color=_color)

        if rn_up is not None and rn_down is not None:
            [ax1.axhline(y=_rn, color=_color, linewidth=0.5) for _rn in rn_up.reshape(-1, 1)]
            [ax1.axhline(y=_rn, color=_color, linewidth=0.5) for _rn in rn_down.reshape(-1, 1)]

        # add detected peaks
        _max_peak_amp = -np.inf
        for i, _ch in enumerate(_channels):
            _current_offset = np.atleast_1d(offset_vector[_offset_idx])[i]
            if all_markers is not None:
                _subset = all_markers.query('channel == "{:}"'.format(_ch))
                for _idx, _m in _subset.iterrows():
                    ax1.add_patch(Rectangle((_m.x_ini.value, _m.y_ini.value - _current_offset),
                                            _m.x_end.value - _m.x_ini.value, _m.y_end.value - _m.y_ini.value,
                                            edgecolor='k'))
            if all_peaks is not None and all_peaks.shape[0] and \
                    not np.all(all_peaks.amp.apply(lambda x_value: np.isnan(x_value.value))):
                _idx_amps = all_peaks.amp.apply(lambda x_value: not np.isnan(x_value.value))
                all_peaks = all_peaks[_idx_amps]
                _max_peak_amp = np.maximum(all_peaks.amp, _max_peak_amp)
                _subset = all_peaks.query('channel == "{:}"'.format(_ch))
                for _idx, _peak in _subset.iterrows():
                    _peak_color = peak_color[np.argwhere(_unique_peaks == _peak.peak_label).squeeze() +
                                             _unique_peaks.size * _idx_data]
                    markerfacecolor = _peak_color if _peak.significant else 'white'
                    x = _peak.x
                    y = _peak.amp
                    if _peak.show_peak and show_peaks:
                        if _peak.positive:
                            ax1.plot(x, y - _current_offset, '^', markersize=3,
                                     markerfacecolor=markerfacecolor, color=_peak_color)
                        else:
                            ax1.plot(x, y - _current_offset, 'v', markersize=3,
                                     markerfacecolor=markerfacecolor, color=_peak_color)
                    if _peak.show_label and show_labels:
                        ax1.text(x, y * 1.1 - _current_offset, _peak.peak_label,
                                 horizontalalignment='center',
                                 verticalalignment='bottom')
                    if _ave_data.domain == Domain.frequency:
                        _pol_ax_idx = i
                        ax_in[_pol_ax_idx].plot(
                            _peak.spectral_phase,
                            y,
                            color=_peak_color,
                            marker='o',
                            linestyle=None,
                            markersize=2,
                            alpha=1,
                            label='')
                        _amp = _max_peak_amp.values[0]
                        ax_in[_pol_ax_idx].set_ylim(0 * _amp.unit, _amp * 1.1)
                        ax_in[_pol_ax_idx].set_yticklabels([])
                        ax_in[_pol_ax_idx].set_xticklabels([])
                        ax_in[_pol_ax_idx].grid(True)
                        ax_in[_pol_ax_idx].set_xlabel('')
                        ax_in[_pol_ax_idx].set_ylabel('')
                        ax_in[_pol_ax_idx].patch.set_alpha(1)

            if show_following_stats is not None:
                all_stats = None
                if statistical_test is not None:
                    if statistical_test in _ave_data.statistical_tests.table_names():
                        all_stats = _ave_data.statistical_tests.get_data(table_name=statistical_test)
                    else:
                        print('Available statistical tests: {:}'.format(_ave_data.statistical_tests.table_names()))
                        print('Could not find statistical test called {:}'.format(statistical_test))

                if all_stats is None:
                    continue
                columns_found = np.intersect1d(all_stats.columns, show_following_stats)
                if columns_found.size == 0:
                    print('Could not find columns: {:}'.format(show_following_stats))
                    continue
                # resort columns to keep input order
                _cols = []
                for _ss in show_following_stats:

                    if _ss in columns_found:
                        _cols.append(_ss)
                columns_found = _cols
                _subset = all_stats.query('channel == "{:}"'.format(_ch))
                _text_stats = ''
                for _idx, _test in _subset.iterrows():
                    for _col in columns_found:
                        _value = _test[_col]
                        if isinstance(_value, str):
                            _text_stats = _text_stats + '{:} '.format(_test[_col])
                        else:
                            _text_stats = _text_stats + '{:} = {:.3f} '.format(_col, _test[_col])
                    _text_stats = _text_stats + '\n'
                my_stats_text.append(_text_stats)
                my_color_text.append(_color)
                my_offset_text.append(_current_offset)
        if _ave_data.roi_windows is not None:
            for _roi in _ave_data.roi_windows:
                if _roi.show_window and not np.isinf(_roi.end_time):
                    ax1.axvspan(_roi.ini_time, _roi.end_time, alpha=0.05, color=_color)
                if _roi.show_label:
                    _y_label = np.max(ax1.get_ylim())
                    ax1.text((_roi.end_time - _roi.ini_time) / 2,
                             _y_label, _roi.label,
                             horizontalalignment='center',
                             verticalalignment='bottom')
        if _idx_data == 0:
            if x_lim is None:
                x_lim = [0, _ave_data.x[-1].value]
            _x = np.min(x_lim)
            ax1.set_title(title, fontsize=fontsize)
            ax1.set_xlim(x_lim)
            if y_lim is not None:
                # ax1.set_ylim(-(n_ch_to_plot + 1) * offset_step, offset_step)
                # else:
                ax1.set_ylim(y_lim)
            y_unit = _ave_data.data.unit
            if y_unit == u.dimensionless_unscaled:
                y_unit = 'A.U.'
            ax1.set_xlabel(x_label,
                           fontsize=fontsize)
            ax1.set_ylabel('Amplitude [{:}]'.format(y_unit),
                           fontsize=fontsize)
            ax2 = ax1.twinx()
            ax2.set_ylim(ax1.get_ylim())
            ax2.set_yticks(-offset_vector)
            ax2.set_yticklabels([_ch for _ch in all_labels], fontsize=fontsize)
            ax2.spines["right"].set_position(("axes", - 0.25))

        text_width = 0

        for _t, _c, _o in zip(my_stats_text, my_color_text, my_offset_text):
            if my_stats_text != '':
                ax_text = ax1.text(_x, -_o, _t,
                                   horizontalalignment='left',
                                   verticalalignment='bottom',
                                   color=_c,
                                   fontsize=6)
                transf = ax1.transData.inverted()
                bb = ax_text.get_window_extent(renderer=fig.canvas.get_renderer())
                bb = bb.transformed(transf)
                text_width = np.maximum(text_width, (bb.xmax - bb.xmin) * 1.1)
        _x = _x + text_width

    # fig.tight_layout()
    if save_to_file:
        _fig_path = figure_dir_path + figure_basename + fig_format
        fig.savefig(_fig_path)
        print('time plots saved figure: {:}'.format(_fig_path))
    return fig


def plot_time_peak_topographic_map(ave_data=DataNode(),
                                   file_name='',
                                   title='',
                                   channel_label=[],
                                   x_lim=None,
                                   y_lim=None,
                                   fontsize=8):
    if x_lim is None:
        x_lim = [ave_data.x[0], ave_data.x[-1]]
    grid_size = 500j
    peak_scalp_potentials = []

    _ch_idx = [_i for _i, _ch in enumerate(ave_data.layout) if _ch.x is not None and _ch.y is not None]
    assigned_channels = ave_data.layout[_ch_idx]

    if not channel_label:
        channel_label = assigned_channels[0].label

    dummy_var = np.array([_i for _i, _ch in enumerate(assigned_channels) if _ch.label == channel_label])
    if not len(dummy_var):
        return
    idx_ch = ave_data.get_channel_idx_by_label([channel_label])

    # check if channels has maximum snr to add in title
    max_snr_idx = np.argmax(ave_data.get_max_snr_per_channel()[_ch_idx])
    if idx_ch == max_snr_idx:
        title += ' best snr'
    # filter peaks for specific channel
    _peaks, _time_windows = get_channel_peaks_and_windows(eeg_peaks=ave_data.peak_times,
                                                          eeg_time_windows=ave_data.peak_time_windows,
                                                          channel_label=channel_label)
    sorted_peaks = None
    _y_max = None
    _y_min = None
    if _peaks:
        # sort peaks by time
        _s_p = itemgetter(*np.argsort([_p.x for _p in _peaks]))(_peaks)
        sorted_peaks = _s_p if isinstance(_s_p, tuple) else (_s_p,)

        max_amps = [_p.amp for _p in ave_data.peak_times if _p.peak_label]
        if max_amps:
            _y_max = np.max(max_amps)
            _y_min = np.min([_p.amp for _p in ave_data.peak_times if _p.peak_label])

        # compute topographic map
        for _peak in sorted_peaks:
            if _peak.peak_label:
                peak_potentials, max_distance = get_potential_fields(epochs_ave=ave_data,
                                                                     x_val=_peak.x,
                                                                     grid=grid_size,
                                                                     domain=_peak.domain)
                # print _peak.peak_label
                # get_CDF(epochs_ave=ave_data,
                #         x_val=_peak.x,
                #         grid=grid_size,
                #         domain=_peak.domain)
                peak_scalp_potentials.append({'potential': peak_potentials,
                                              'peak': _peak.peak_label})

    # plot topographic map
    inch = 2.54
    fig = plt.figure()
    fig.set_size_inches(18 / inch, 12 / inch)

    row_idx = 0
    gs = gridspec.GridSpec(2, len(peak_scalp_potentials)) if peak_scalp_potentials else gridspec.GridSpec(1, 1)
    if peak_scalp_potentials:
        row_idx += 1
        for _idx, peak_field in enumerate(peak_scalp_potentials):
            ax = plt.subplot(gs[0, _idx])
            ax_im = ax.imshow(peak_field['potential'].T, origin='lower',
                              extent=(-max_distance, max_distance, -max_distance, max_distance),
                              vmin=_y_min,
                              vmax=_y_max,
                              aspect=1.0)
            ax_im.set_cmap('nipy_spectral')
            levels = np.arange(_y_min, _y_max, (_y_max - _y_min) / 5.0)
            ax.contour(peak_field['potential'].T,
                       levels,
                       origin='lower',
                       extent=(-max_distance, max_distance, -max_distance, max_distance),
                       linewidths=1.0,
                       colors='k')
            ax.autoscale(enable=False)
            ax.plot(0, max_distance * 1.0, '|', markersize=5, color='k')
            for i, _ch in enumerate(ave_data.layout):
                if not (_ch.x is None or _ch.y is None):
                    ax.plot(_ch.x, _ch.y, 'o', color='b', markersize=0.2)
            ax.get_xaxis().set_ticks([])
            ax.get_yaxis().set_ticks([])
            ax.axis('off')
            ax.set_title(peak_field['peak'])

        c_bar_ax = fig.add_axes([0.05, 0.6, 0.01, 0.25])
        fig.colorbar(ax_im, cax=c_bar_ax, orientation='vertical', format='%.1f')
        c_bar_ax.yaxis.set_ticks_position('left')

    ax2 = plt.subplot(gs[row_idx, 0:])
    ax2.plot(ave_data.x, ave_data.data[:, idx_ch])
    ax2.axhline(y=ave_data.rn[idx_ch], color='k', linewidth=0.3)
    ax2.axhline(y=-ave_data.rn[idx_ch], color='k', linewidth=0.3)
    if sorted_peaks is not None:
        for _idx, _peak in enumerate(sorted_peaks):
            if _peak.peak_label and _peak.show_label:
                markerfacecolor = 'black' if _peak.significant else 'white'
                if _peak.positive:
                    ax2.plot(_peak.x, _peak.amp, '^', markersize=3, markerfacecolor=markerfacecolor)
                else:
                    ax2.plot(_peak.x, _peak.amp, 'v', markersize=3, markerfacecolor=markerfacecolor)
                ax2.text(_peak.x, _peak.amp * 1.1, _peak.peak_label, horizontalalignment='center',
                         verticalalignment='bottom')
    ax2.set_xlabel('Time [{:}]'.format(ave_data.x.unit), fontsize=fontsize)
    ax2.set_xlim(x_lim)
    ax2.set_ylabel('Amplitude [{:}]'.format(ave_data.data.unit), fontsize=fontsize)
    if _y_max and _y_min and y_lim is None:
        ax2.set_ylim([_y_min * 1.2, (_y_max + (_y_max - _y_min) / 10.0) * 1.2])
    elif y_lim is not None:
        ax2.set_ylim(y_lim)
    ax2.set_title(title, fontsize=fontsize)
    if _time_windows:
        [ax2.axvspan(_t_w.ini_time, _t_w.end_time, alpha=0.15, color='r')
         for _t_w in _time_windows if _t_w.positive_peak and _t_w.show_window]
        [ax2.axvspan(_t_w.ini_time, _t_w.end_time, alpha=0.15, color='b')
         for _t_w in _time_windows if not _t_w.positive_peak and _t_w.show_window]
    plt.tight_layout()
    fig.savefig(file_name)
    plt.close(fig)
    gc.collect()
    print(('figure saved: ' + file_name))


def plot_time_topographic_map(ave_data=DataNode(),
                              title: str = '',
                              times: u.Quantity = np.array([]),
                              channel_label: [str] = None,
                              x_lim: [float, float] = None,
                              y_lim: [float, float] = None,
                              fontsize: float = 8):
    if x_lim is None:
        x_lim = [ave_data.x[0], ave_data.x[-1]]
    grid_size = 500j
    peak_scalp_potentials = []
    times = set_default_unit(times, u.s)
    # _ch_idx = [_i for _i, _ch in enumerate(ave_data.layout) if _ch.x is not None and _ch.y is not None]
    # assigned_channels = ave_data.layout[_ch_idx]
    # if channel_label is None:
    #     channel_label = assigned_channels[0].label
    #
    # dummy_var = np.array([_i for _i, _ch in enumerate(assigned_channels) if _ch.label == channel_label])
    # if not len(dummy_var):
    #     return
    idx_ch = ave_data.get_channel_idx_by_label([channel_label])
    if not idx_ch.size:
        return

    # check if channels has maximum snr to add in title
    # max_snr_idx = np.argmax(ave_data.get_max_snr_per_channel()[_ch_idx])
    # if idx_ch == max_snr_idx:
    #     title += ' best snr'

    _y_max = None
    _y_min = None
    _time_labels = np.array([''] * times.size)
    _show_labels = np.array([True] * times.size)
    if ave_data.peaks is not None and ave_data.peaks.shape[0]:
        _subset = ave_data.peaks.query('channel == "{:}"'.format(channel_label))
        _subset = _subset.reset_index(drop=True)
        if _subset.shape[0]:
            # sort peaks by time
            _time_points = ave_data.x_to_samples(_subset['x'])
            _original_idx = np.argsort(_time_points)
            _peak_labels = _subset['peak_label'][_original_idx]
            _show = _subset['show_label'][_original_idx]
            times = np.append(times, ave_data.x[_time_points[_original_idx]])
            _time_labels = np.append(_time_labels, _peak_labels)
            _show_labels = np.append(_show_labels, _show)

    if times.size:
        amps = ave_data.data[ave_data.x_to_samples(times), :]
        _y_max = np.nanmax(amps)
        _y_min = np.nanmin(amps)

        # compute topographic map
        max_distance = None
        for _idx_t, _t in enumerate(times):
            peak_potentials, max_distance = get_potential_fields(epochs_ave=ave_data,
                                                                 x_val=_t,
                                                                 grid=grid_size,
                                                                 domain=ave_data.domain)
            if peak_potentials is None or not peak_potentials.size:
                continue
            _label = _time_labels[_idx_t] if _show_labels[_idx_t] else ''
            peak_scalp_potentials.append({'potential': peak_potentials,
                                          'peak': '{:} {:.2e}'.format(_label, _t)}
                                         )

        if len(peak_scalp_potentials) == 0:
            return
        # plot topographic map
        inch = 2.54
        fig = plt.figure()
        fig.set_size_inches(18 / inch, 12 / inch)

        row_idx = 0
        gs = gridspec.GridSpec(2, len(peak_scalp_potentials)) if peak_scalp_potentials else gridspec.GridSpec(1, 1)
        if peak_scalp_potentials:
            row_idx += 1
            for _idx, peak_field in enumerate(peak_scalp_potentials):
                ax = plt.subplot(gs[0, _idx])
                ax_im = ax.imshow(peak_field['potential'].T, origin='lower',
                                  extent=(-max_distance, max_distance, -max_distance, max_distance),
                                  vmin=_y_min.value,
                                  vmax=_y_max.value,
                                  aspect=1.0)
                ax_im.set_cmap('nipy_spectral')
                levels = np.arange(_y_min.value, _y_max.value, (_y_max.value - _y_min.value) / 5.0)
                ax.contour(peak_field['potential'].T,
                           levels,
                           origin='lower',
                           extent=(-max_distance, max_distance, -max_distance, max_distance),
                           linewidths=1.0,
                           colors='k')
                ax.autoscale(enable=False)
                ax.plot(0, max_distance * 1.0, '|', markersize=5, color='k')
                for i, _ch in enumerate(ave_data.layout):
                    if not (_ch.x is None or _ch.y is None):
                        ax.plot(_ch.x, _ch.y, 'o', color='b', markersize=0.2)
                ax.get_xaxis().set_ticks([])
                ax.get_yaxis().set_ticks([])
                ax.axis('off')
                ax.set_title(peak_field['peak'])

            c_bar_ax = fig.add_axes([0.05, 0.6, 0.01, 0.25])
            fig.colorbar(ax_im, cax=c_bar_ax, orientation='vertical', format='%.1f')
            c_bar_ax.yaxis.set_ticks_position('left')

        ax2 = plt.subplot(gs[row_idx, 0:])
        ax2.plot(ave_data.x, ave_data.data[:, idx_ch])
        ax2.axhline(y=ave_data.rn[idx_ch].to_value(ave_data.data.unit), color='k', linewidth=0.3)
        ax2.axhline(y=-ave_data.rn[idx_ch].to_value(ave_data.data.unit), color='k', linewidth=0.3)
        for _t in times:
            ax2.axvline(_t)
        ax2.set_xlabel('Time [{:}]'.format(ave_data.x.unit), fontsize=fontsize)
        ax2.set_xlim(x_lim)
        ax2.set_xticks(np.linspace(min(x_lim), max(x_lim), num=10))  # sets xticks
        _unit = ave_data.data.unit
        if _unit == u.dimensionless_unscaled:
            _unit = 'A.U.'
        ax2.set_ylabel('Amplitude [{:}]'.format(_unit), fontsize=fontsize)
        if _y_max is not None and _y_min is not None and y_lim is None:
            ax2.set_ylim([_y_min * 1.2, (_y_max + (_y_max - _y_min) / 10.0) * 1.2])
        elif y_lim is not None:
            ax2.set_ylim(y_lim)
        ax2.set_title(title, fontsize=fontsize)
        return fig


def plot_freq_topographic_map(ave_data=DataNode(),
                              title='',
                              channel_label: [str] = None,
                              x_lim=None,
                              y_lim=None,
                              fontsize=8.0):
    plt.ioff()
    if x_lim is None:
        x_lim = [0, ave_data.x[-1].value]
    grid_size = 150j
    peak_scalp_potentials = []

    idx_ch = ave_data.get_channel_idx_by_label([channel_label])
    if idx_ch.size == 0:
        return
    _peaks, _ = get_channel_peaks_and_windows(eeg_peaks=ave_data.peaks,
                                              channel_label=channel_label)

    sorted_peaks = None
    _y_max = None
    max_distance = None
    if _peaks.size:
        # sort peaks by frequency
        sorted_peaks = _peaks.sort_values('x')

        _y_max = ave_data.peaks['amp'].max()

        # compute topographic map
        for _, _peak in sorted_peaks.iterrows():
            if _peak.peak_label:
                peak_potentials, max_distance = get_potential_fields(epochs_ave=ave_data,
                                                                     x_val=_peak.x,
                                                                     grid=grid_size,
                                                                     domain=_peak.domain)
                if peak_potentials is not None:
                    peak_scalp_potentials.append({'potential': peak_potentials,
                                                  'peak': _peak.peak_label})

    # plot topographic map
    inch = 2.54
    fig = plt.figure()
    fig.set_size_inches(18 / inch, 12 / inch)
    row_idx = 0
    gs = gridspec.GridSpec(2, len(peak_scalp_potentials)) if peak_scalp_potentials else gridspec.GridSpec(1, 1)

    if peak_scalp_potentials:
        row_idx += 1
        for _idx, peak_field in enumerate(peak_scalp_potentials):
            ax = plt.subplot(gs[0, _idx])
            ax_im = ax.imshow(peak_field['potential'].T, origin='lower',
                              extent=(-max_distance, max_distance, -max_distance, max_distance),
                              vmin=0,
                              vmax=_y_max.value,
                              aspect=1.0)
            ax_im.set_cmap('nipy_spectral')
            levels = np.arange(0, 5) * _y_max / 4
            ax.contour(peak_field['potential'].T,
                       levels,
                       origin='lower',
                       extent=(-max_distance, max_distance, -max_distance, max_distance),
                       linewidths=1.0,
                       colors='k')
            ax.autoscale(enable=False)
            ax.plot(0, max_distance * 1.0, '|', markersize=5, color='k')
            for i, _ch in enumerate(ave_data.layout):
                if not (_ch.x is None or _ch.y is None):
                    ax.plot(_ch.x, _ch.y, 'o', color='b', markersize=0.2)
            ax.get_xaxis().set_ticks([])
            ax.get_yaxis().set_ticks([])
            ax.axis('off')
            ax.set_title(peak_field['peak'])

        c_bar_ax = fig.add_axes([0.05, 0.6, 0.01, 0.25])
        fig.colorbar(ax_im, cax=c_bar_ax, orientation='vertical', format='%.1f')
        c_bar_ax.yaxis.set_ticks_position('left')

    # plot waveform of channel used
    ax2 = plt.subplot(gs[row_idx, 0:])
    ax2.plot(ave_data.x, np.abs(ave_data.data[:, idx_ch]))
    ax2.set_xlim(x_lim)
    if sorted_peaks is not None:
        for _idx, _peak in sorted_peaks.iterrows():
            if _peak.show_label:
                markerfacecolor = 'black' if _peak.significant else 'white'
                if _peak.positive:
                    ax2.plot(_peak.x,
                             _peak.amp,
                             'v',
                             markersize=3,
                             markerfacecolor=markerfacecolor)
                else:
                    ax2.plot(_peak.x,
                             _peak.amp,
                             '^',
                             markersize=3,
                             markerfacecolor=markerfacecolor)
                if _peak.peak_label:
                    ax2.text(_peak.x,
                             _peak.amp + 0.1 * _peak.amp.unit,
                             _peak.peak_label,
                             horizontalalignment='center',
                             verticalalignment='bottom',
                             fontsize=fontsize)
    ax2.set_xlabel('Frequency [{:}]'.format(u.Hz),
                   fontsize=fontsize)
    _unit = ave_data.data.unit
    if _unit == u.dimensionless_unscaled:
        _unit = 'A.U.'
    ax2.set_ylabel('Amplitude [{:}]'.format(_unit), fontsize=fontsize)

    if _y_max is not None and y_lim is None:
        ax2.set_ylim([0 * _y_max.unit, _y_max * 1.2])
    elif y_lim is not None:
        ax2.set_ylim(y_lim)
    ax2.set_title(title, fontsize=fontsize)
    return fig


def plot_eeg_time_frequency_power(ave_data: DataNode = DataNode(),
                                  figure_dir_path: str = '',
                                  figure_basename: str = '',
                                  eeg_topographic_map_channels: np.array = np.array([]),
                                  title: str = None,
                                  x_lim: [float, float] = None,
                                  y_lim: [float, float] = None,
                                  fig_format: str = '.pdf',
                                  fontsize: float = 8,
                                  return_figures: bool = True,
                                  save_figures: bool = False,
                                  normalize: bool = True,
                                  db_scale: bool = True,
                                  spec_thresh: float = -np.inf
                                  ):
    if eeg_topographic_map_channels.size:
        topographic_map_channels = eeg_topographic_map_channels
    else:
        topographic_map_channels = [_ch.label for _ch in ave_data.layout]
    figures = []
    for _label in topographic_map_channels:
        _fig_path = None
        if save_figures:
            _fig_path = figure_dir_path + '_' + _label + '_' + figure_basename + '_power_spectrogram_.' + fig_format
        _title = ''
        if title is not None and len(title) > 0:
            _title = ' / ' + title
        fig = plot_time_frequency_power_transformation(ave_data=ave_data,
                                                       channel_label=_label,
                                                       file_name=_fig_path,
                                                       title=_label + _title,
                                                       x_lim=x_lim,
                                                       y_lim=y_lim,
                                                       fontsize=fontsize,
                                                       save_figures=save_figures,
                                                       spec_thresh=spec_thresh,
                                                       normalize=normalize,
                                                       db_scale=db_scale)
        if return_figures:
            figures.append(fig)
        else:
            plt.close(fig)
            gc.collect()
    return figures


def plot_time_frequency_power_transformation(ave_data: DataNode = DataNode(),
                                             file_name='',
                                             title='',
                                             channel_label: [str] = None,
                                             x_lim=None,
                                             y_lim=None,
                                             fontsize=8,
                                             save_figures: bool = False,
                                             spec_thresh: float = -np.inf,
                                             normalize: bool = True,
                                             db_scale: bool = True):

    if x_lim is None:
        x_lim = [ave_data.x[0], ave_data.x[-1]]

    _ch_idx = [_i for _i, _ch in enumerate(ave_data.layout) if _ch.x is not None and _ch.y is not None]
    assigned_channels = ave_data.layout[_ch_idx]

    if channel_label is None:
        channel_label = assigned_channels[0].label

    dummy_var = np.array([_i for _i, _ch in enumerate(assigned_channels) if _ch.label == channel_label])
    if not len(dummy_var):
        return
    idx_ch = ave_data.get_channel_idx_by_label([channel_label])

    if not len(idx_ch):
        return

    # plot topographic map
    inch = 2.54
    fig = plt.figure()
    fig.set_size_inches(18 / inch, 12 / inch)

    row_idx = 0
    gs = gridspec.GridSpec(1, 1)
    ax2 = plt.subplot(gs[row_idx, 0:])

    power = np.squeeze(np.abs(ave_data.data[:, idx_ch, :]))
    if normalize:
        power /= power.max()
    if db_scale:
        power = 10 * np.log10(power)
    power[power < spec_thresh] = spec_thresh
    ax_imag = ax2.pcolormesh(ave_data.x,
                             ave_data.y,
                             power.T,
                             cmap=plt.get_cmap('jet'))

    ax2.set_ylabel('Frequency [{:}]'.format(ave_data.y.unit))
    ax2.set_xlabel('Time [{:}]'.format(ave_data.x.unit), fontsize=fontsize)
    ax2.set_xlim(x_lim)
    if y_lim is not None:
        ax2.set_ylim(y_lim)
    ax2.set_title(title, fontsize=fontsize)
    a_pos = ax2.get_position()
    cbaxes = fig.add_axes([a_pos.x1 + 0.005, a_pos.y0, 0.005, a_pos.height])
    color_map_label = 'Amplitude [{:}]'.format(ave_data.data.unit)
    if db_scale:
        color_map_label = 'Amplitude [dB]'
    if db_scale and normalize:
        color_map_label = 'Amplitude [dB ref max]'
    c_bar = fig.colorbar(ax_imag, orientation='vertical', format='%.1f', cax=cbaxes)
    c_bar.set_label(color_map_label, fontsize=fontsize)
    if save_figures:
        fig.savefig(file_name)
        print(('figure saved: ' + file_name))
    return fig


def plot_eeg_topographic_map(ave_data: DataNode = DataNode(),
                             figure_dir_path: str = '',
                             figure_basename: str = '',
                             return_figures: bool = False,
                             eeg_topographic_map_channels: np.array = np.array([]),
                             domain: Domain = Domain.time,
                             save_figures: bool = True,
                             times: np.array = None,
                             title: str = '',
                             x_lim: [] = None,
                             y_lim: [] = None,
                             fig_format: str = '.pdf',
                             fontsize: float = 8) -> [plt.figure]:
    """
    This function produces a matplotlib figure with the topographic map of the input DataNode.
    :param ave_data: DataNode containing the data to be plotted (times x channels)
    :param figure_dir_path: path where figures will be saved
    :param figure_basename: this is the base file name to save figures. Since the output will consist of several
    figures (one per channel), the channel label will be appended to the figure_basename
    :param return_figures: if True, the matplotlib figure will be returned
    :param eeg_topographic_map_channels: A list of strings with the labels of the channels to will be plotted. If left
    empty, all channels will be plotted.
    :param domain: the domain of the topographic map (Domain.time or Domain.frequency)
    :param save_figures: if True, figures will be saved on the figure_dir_path
    :param times: numpy array with times that will be used to generate the respective topographic maps.
    :param title: string with the title of the figure
    :param x_lim: numerical list with the range of the x-axis (min and max).
    :param y_lim: numerical list with the range of the y-axis (min and max).
    :param fig_format: the desired format of the saved figures (e.g. '.png', '.pdf')
    :param fontsize: float indicating the font size (in points) used to render text in the figure
    :return: list of figures
    """
    times = set_default_unit(times, u.s)
    if times is not None and times.ndim == 0:
        times = np.array([times.to_value()]) * times.unit
    eeg_topographic_map_channels = np.array(eeg_topographic_map_channels)
    figures = []
    if eeg_topographic_map_channels.size:
        topographic_map_channels = eeg_topographic_map_channels
    else:
        topographic_map_channels = [_ch.label for _ch in ave_data.layout]

    for _label in topographic_map_channels:
        _title = _label if title == '' else _label + ' / ' + title
        if domain == Domain.time:
            _fig_path = os.path.join(figure_dir_path, _label + '_' + figure_basename + '_TMap' + fig_format)
            fig = plot_time_topographic_map(ave_data=ave_data,
                                            channel_label=_label,
                                            times=times,
                                            title=_title,
                                            x_lim=x_lim,
                                            y_lim=y_lim,
                                            fontsize=fontsize)
            if fig is not None and save_figures:
                fig.savefig(_fig_path)
                print('figure saved: {:}'.format(_fig_path))
            if not return_figures:
                plt.close(fig)
                gc.collect()
            else:
                figures.append(fig)
        if domain == Domain.frequency:
            _fig_path = os.path.join(figure_dir_path, _label + '_' + figure_basename + '_TMap_freq' + fig_format)
            fig = plot_freq_topographic_map(ave_data=ave_data,
                                            channel_label=_label,
                                            title=_title,
                                            x_lim=x_lim,
                                            y_lim=y_lim,
                                            fontsize=fontsize)
            if fig is not None and save_figures:
                fig.savefig(_fig_path)
                print('figure saved: {:}'.format(_fig_path))
            if not return_figures:
                plt.close(fig)
                gc.collect()
            else:
                figures.append(fig)
    return figures


def eeg_save_time_slice(data: np.array = np.array([]),
                        fs: float = 0,
                        channels: np.array([EegChannel]) = np.array([EegChannel()]),
                        time_length: float = 4.0,
                        title: str = '',
                        file_name: str = ''):
    buffer_size = np.floor(np.minimum(time_length * fs, data.shape[0])).astype(int)
    fig, ax1 = plt.subplots()
    font = {'size': 6}
    plt.rc('font', **font)
    offset_vector = 10 * np.arange(data.shape[1])
    ax1.plot(np.arange(buffer_size) / fs, data[np.arange(buffer_size), :] - offset_vector,
             linewidth=0.5)
    # if interpolate_data:
    #     ax1.plot(interpolation_data_points['ini'],
    #              np.ones((interpolation_data_points['ini'].shape[0], data.shape[1]))
    #              + np.mean(data[np.arange(buffer_size), :], axis=0) - offset_vector,
    #              marker='v',
    #              linestyle='None',
    #              markerfacecolor='black',
    #              markersize=1.0)
    #
    #     ax1.plot(interpolation_data_points['end'],
    #              np.ones((interpolation_data_points['end'].shape[0], data.shape[1]))
    #              + np.mean(data[np.arange(buffer_size), :], axis=0) - offset_vector,
    #              marker='^',
    #              linestyle='None',
    #              markerfacecolor='black',
    #              markersize=1.0)
    ax1.set_title(title)
    ax1.set_xlabel('Time [s]')
    ax1.set_ylabel('Amplitude [' + r'$\mu$' + 'V]')
    ax2 = ax1.twinx()
    ax2.set_ylim(ax1.get_ylim())
    ax2.set_yticks(np.mean(data[np.arange(buffer_size), :] - offset_vector, axis=0))
    ax2.set_yticklabels([ch.label for i, ch in enumerate(channels)])
    ax2.spines["right"].set_position(("axes", - 0.1))
    fig.savefig(file_name)
    plt.close(fig)
    gc.collect()


def get_cdf(epochs_ave: DataNode = DataNode(),
            x_val: float = None,
            domain=Domain.time,
            **kwargs):
    idx_elec = np.where(np.array([ch.x for ch in epochs_ave.channels]))[0]
    # remove electrodes without a label
    if domain == Domain.time:
        potentials = epochs_ave.data[epochs_ave.time_to_samples(x_val), idx_elec]
    if domain == Domain.frequency:
        potentials = epochs_ave.rfft_average[epochs_ave.frequency_to_samples(x_val), idx_elec]

    x = np.zeros((potentials.size, 1))
    y = np.zeros((potentials.size, 1))
    z = np.zeros((potentials.size, 1))
    elec_pos = np.zeros((potentials.size, 2))
    _channels = np.array(epochs_ave.channels)[idx_elec]
    for i, _ch in enumerate(_channels):
        if _ch.x is not None and _ch.y is not None:
            x[i] = _ch.x
            y[i] = _ch.y
            z[i] = potentials[i]
            elec_pos[i] = np.array([_ch.x, _ch.y])
    # params = {'gdX': 0.05,
    #           'gdY': 0.05,
    #           'gdZ': 0.05,
    #           'n_sources': 64
    #           }
    # elec_pos = get_3D_spherical_positions(x, y)

    # params = {'gdX': 0.05,
    #           'gdY': 0.05,
    #           'n_sources': 64
    #           }

    # k = KCSD(elec_pos, z, params)
    #
    # k.estimate_pots()
    # k.estimate_csd()

    # k.plot_all()
    return interpolate_potential_fields(x, y, z, **kwargs)



"""
.. _tut-acc-sim:

Test spatial filtering with evoked response (Simulation)
==============================

This example simulates an evoked responses with non-stationary noise.
We illustrate how weighted averaging and DSS can lead to much better results in terms of recovering the data.
This example follows that by Alain de Cheveigné (http://audition.ens.fr/adc/NoiseTools/) but it has been extended
to include non-stationary noise and weighted average.

"""
import matplotlib.pyplot as plt
import peegy.processing.tools.filters.spatial_filtering.spatial_filtering as sf
from peegy.processing.tools.eeg_epoch_operators import et_mmat, et_freq_shifted_xcovariance, \
    et_freq_weighted_cov
from peegy.processing.tools.epochs_processing_tools import et_mean
import numpy as np
import matplotlib.gridspec as gridspec
from scipy.stats import norm
__author__ = 'jundurraga'


def plot_covariances(covariances: [np.array] = None,
                     normalized=True):
    gs = gridspec.GridSpec(1, len(covariances))
    fig = plt.figure()
    for _i, _cov in enumerate(covariances):
        if normalized:
            _cxy = _cov[np.diag_indices_from(_cov)][None, :] ** 0.5
            _norm_cxy = _cxy.T.dot(_cxy)
            _cov = _cov / _norm_cxy
        ax = plt.subplot(gs[0, _i])
        maxis = ax.matshow(_cov)
        plt.colorbar(maxis)
    return fig


def plot_data_results(data_list: [np.array] = None, titles: [str] = None):
    gs = gridspec.GridSpec(1, len(data_list))
    fig = plt.figure()
    for _i, (_data, _title) in enumerate(zip(data_list, titles)):
        ax = plt.subplot(gs[0, _i])
        ax.plot(_data)
        ax.set_title(_title)
    return fig


# create synthetic data
nsamples = 500 * 3
nchans = 30
ntrials = 300
noise_dim = 20  # dimensionality of noise
freq = 100
fs = 1000
source = np.concatenate((
    np.zeros((nsamples // 3, )),
    np.sin(2 * np.pi * np.arange(nsamples / 3) / fs * freq),
    np.zeros((nsamples // 3, ))))

# source = np.sin(2 * np.pi * np.arange(nsamples) / fs * freq)

np.random.seed(2)
mix_m = norm.ppf(np.random.rand(nchans, 1))
s = source * mix_m
s = np.tile(np.expand_dims(s.T, axis=2), (1, 1, ntrials))
SNR = 0.5

noise_unmixed = norm.ppf(np.random.rand(noise_dim * ntrials * nsamples))
noise_unmixed = np.reshape(noise_unmixed, [nsamples, noise_dim, ntrials], order='F')

noise_mix = norm.ppf(np.random.rand(nchans * noise_dim))
noise_mix = np.reshape(noise_mix, [noise_dim, nchans], order='F')
noise = et_mmat(noise_unmixed, noise_mix)

scaled_source = SNR * s / np.std(s)
scaled_noise = noise / np.std(noise)

# generate non-stationary noise
for i in np.arange(1, 4):
    _ini = i*20
    _end = _ini + 10
    scaled_noise[:, :, _ini: _end] += 1e2 * i * scaled_noise[:, :, _ini: _end]

# mix data
data = (scaled_noise + scaled_source)

# standard averaging of the data
s_ave, s_w, s_rn, *_ = et_mean(data, block_size=1, weighted=False)
# weighted averaging of the data
w_ave, w_w, w_rn, *_ = et_mean(data, block_size=1, weighted=True, weight_across_epochs=False)
n_components = np.arange(1)

# clean data using weighted dss
n_freq = freq * np.arange(4) / fs
cref, tref = et_freq_weighted_cov(data, normalized_frequencies=n_freq)
c1, t1 = et_freq_weighted_cov(w_ave, normalized_frequencies=n_freq)
cref = cref / tref
c1 = c1 / t1
todss, pwr0, pwr1, n_0, n_1 = sf.nt_dss0(cref, c1, keep1=1e-5)
z0 = et_mmat(data, todss)
cov_1, tw_cov = et_freq_shifted_xcovariance(z0, data, wy=w_w, normalized_frequencies=n_freq)
cov_1 = cov_1 / tw_cov
data_clean0 = et_mmat(z0[:, n_components, :], cov_1[n_components, :])

# clean data using dss with standard average
cref, tref = et_freq_weighted_cov(data, normalized_frequencies=n_freq, w=s_w)
c1, t1 = et_freq_weighted_cov(s_ave, normalized_frequencies=n_freq)
cref = cref / tref
c1 = c1 / t1
todss, pwr0, pwr1, n_0, n_1 = sf.nt_dss0(cref, c1)
z1 = et_mmat(data, todss)
cov_1, tw_cov = et_freq_shifted_xcovariance(z1, data, wy=s_w, normalized_frequencies=n_freq)
cov_1 = cov_1 / tw_cov
data_clean1 = et_mmat(z1[:, n_components, :], cov_1[n_components, :])


# clean data using dss only on weighted average
cref, tref = et_freq_weighted_cov(data, normalized_frequencies=n_freq, w=s_w)
c1, t1 = et_freq_weighted_cov(w_ave, normalized_frequencies=n_freq)
cref = cref / tref
c1 = c1 / t1
todss, pwr0, pwr1, n_0, n_1 = sf.nt_dss0(cref, c1, keep1=1e-5)
z2 = et_mmat(data, todss)
cov_1, tw_cov = et_freq_shifted_xcovariance(z2, data, normalized_frequencies=n_freq, wy=s_w)
cov_1 = cov_1 / tw_cov
data_clean2 = et_mmat(z2[:, n_components, :], cov_1[n_components, :])

# plot the results

w_ave_0, *_ = et_mean(data_clean0, block_size=5, weighted=True)
w_ave_1, *_ = et_mean(data_clean1, block_size=5, weighted=True)
w_ave_2, *_ = et_mean(data_clean2, block_size=5, weighted=True)
w_ave_z0, *_ = et_mean(z0[:, 0:3, :], block_size=5, weighted=True)
w_ave_z1, *_ = et_mean(z1[:, 0:3, :], block_size=5, weighted=True)
w_ave_z2, *_ = et_mean(z2[:, 0:3, :], block_size=5, weighted=True)
w_ave_z0 = np.abs(np.fft.rfft(w_ave_0, axis=0))
w_ave_z1 = np.abs(np.fft.rfft(w_ave_1, axis=0))
w_ave_z2 = np.abs(np.fft.rfft(w_ave_2, axis=0))
plot_data_results([np.mean(scaled_source, axis=2),
                   s_ave,
                   w_ave,
                   w_ave_0,
                   w_ave_1,
                   w_ave_2,
                   w_ave_z0,
                   w_ave_z1,
                   w_ave_z2],
                  ['Target source',
                   'standard average',
                   'weighted average',
                   'w_cov & w_ave',
                   's_cov & s_ave',
                   's_cov & w_ave',
                   'components data_clean0',
                   'components data_clean1',
                   'components data_clean2',
                   ]
                  )
plt.show()

# plot weights and noise
plt.figure()
noise_sample = scaled_noise[0, 0, :]
plt.plot(noise_sample / np.abs(noise_sample).max())
w_w_sample = w_w[0, 0, :]
plt.plot(w_w_sample / np.max(w_w_sample))
plt.show()



import matplotlib.pyplot as plt
import numpy as np
import peegy.processing.tools.filters.spatial_filtering.spatial_filtering as sf
import peegy.processing.tools.eeg_epoch_operators as eo
__author__ = 'jundurraga-ucl'
# create synthetic data
fs = 44100.0
nsamples = int(np.round(1 * fs))
nchans = 36
ntrials = 100
noise_dim = 36  # dimensionality of noise
f1 = 40
source = np.expand_dims(np.sin(2 * np.pi * f1 * np.arange(nsamples) / fs + 3*np.pi/2), axis=1)
coeff = np.arange(nchans/2) * 0.5 / (nchans / 2)
coeff = np.expand_dims(np.hstack((coeff, -coeff)), 0)

# s = source * np.random.normal(0, 0.1, size=(1, nchans))
s = source * coeff
s = np.tile(np.expand_dims(s, axis=2), (1, 1, ntrials))
SNR = 10
noise = np.random.normal(0, 0.1, size=(nsamples, nchans, ntrials))
n_source = SNR * s / np.std(s)
data = noise / np.std(noise) + n_source

c0, c1 = sf.nt_bias_fft(data, np.array([f1]) / fs)

todss, pwr0, pwr1, n0, n1 = sf.nt_dss0(c0, c1)
z = eo.et_mmat(data, todss)
n_components = np.arange(1)
cov_1 = eo.et_x_covariance(z, data)
data_clean = eo.et_mmat(z[:, n_components, :], cov_1[n_components, :])

ax = plt.subplot(131)
ax.plot(np.mean(n_source, axis=2))
ax = plt.subplot(132)
ax.plot(np.mean(data, axis=2))
ax = plt.subplot(133)
ax.plot(np.mean(data_clean, axis=2))
plt.show()



"""
.. _tut-assr-phase-locking-sim:

########################################################
ASSR Phase Locking Value Test Example (Simulated)
########################################################

In this example we simulate an ASSR, and we investigate the significance of the phase by means of a Rayleigh Test on
the frequency of interest

"""
# Enable below for interactive backend
# import matplotlib
# if 'Qt5Agg' in matplotlib.rcsetup.all_backends:
#    matplotlib.use('Qt5Agg')
from peegy.processing.pipe.pipeline import PipePool
from peegy.processing.pipe.definitions import Events, Domain
from peegy.processing.pipe.general import ReferenceData, FilterData, AutoRemoveBadChannels
from peegy.processing.pipe.epochs import EpochData
from peegy.processing.pipe.attach import AppendGFPChannel
from peegy.processing.pipe.statistics import PhaseLockingValue
from peegy.processing.pipe.plot import PlotTopographicMap
from peegy.processing.pipe.spatial_filtering import CreateAndApplySpatialFilter
from peegy.processing.pipe.simulate import GenerateInputData
from peegy.processing.tools.template_generator.auditory_waveforms import aep
from peegy.processing.pipe.storage import MeasurementInformation, SubjectInformation, SaveToDatabase
import os
import astropy.units as u
import numpy as np


# %%
# Generate some data
# ===================
# We generate some auditory steady-state response (ASSR)


fs = 256.0 * u.Hz
epoch_length = 4.0 * u.s
epoch_length = np.ceil(epoch_length * fs) / fs  # fit to fs rate
assr_frequency = 41.0 * u.Hz
assr_frequency = np.ceil(epoch_length * assr_frequency) / epoch_length  # fit to epoch length
# here we pick some random frequencies to test statistical detection
random_frequencies = np.unique(np.random.rand(10) * 5) * u.Hz
template_waveform, _ = aep(fs=fs)
n_channels = 32
event_times = np.arange(0, 360.0, 1 / assr_frequency.to(u.Hz).value) * u.s
reader = GenerateInputData(template_waveform=template_waveform,
                           fs=fs,
                           n_channels=n_channels,
                           mixing_matrix=np.diag(np.arange(n_channels))/n_channels,
                           snr=0.01,
                           layout_file_name='biosemi32.lay',
                           event_times=event_times,
                           event_code=1.0,
                           figures_subset_folder='assr_plv_rayleigh_test')
reader.run()

# %%
# Resize events
# ========================
# Now we keep events at intervals that correspond to our desired epoch length
events = reader.output_node.events.get_events(code=1)
# skip events to preserve only those at each epoch point
_new_events = Events(events=events[0:-1:int(epoch_length * assr_frequency)])
reader.output_node.events = _new_events

# %%
# Start the pipeline
# ========================
# Some processing to obtain clean epochs


pipeline = PipePool()
pipeline['referenced'] = ReferenceData(reader, reference_channels=['Cz'], invert_polarity=True)
pipeline['channel_cleaned'] = AutoRemoveBadChannels(pipeline['referenced'])
pipeline['time_filtered_data'] = FilterData(pipeline['channel_cleaned'],
                                            high_pass=2.0 * u.Hz,
                                            low_pass=100.0 * u.Hz)
pipeline['time_epochs'] = EpochData(pipeline['time_filtered_data'],
                                    event_code=1.0,
                                    base_line_correction=False)
pipeline.run()

# %%
# Compute and plot components in the frequency-domain
# -----------------------------------------------------
# Spatial filter is a applied in the frequency-domain

pipeline['dss_time_epochs'] = CreateAndApplySpatialFilter(pipeline['time_epochs'],
                                                          test_frequencies=np.array([assr_frequency.to(u.Hz).value]),
                                                          sf_components=np.arange(0, 2),
                                                          components_to_plot=np.arange(0, 5),
                                                          projection_domain=Domain.frequency)
pipeline.run()

# %%
# Compute global field power (GFP)
# ---------------------------------
# We compute the GFP across channels and epochs. A new channel with the GFP is appended to the data


pipeline['dss_time_epochs_with_gfp'] = AppendGFPChannel(pipeline['dss_time_epochs'])
pipeline.run()

# %%
# Run Rayleigh Tests (phase-locking value)
# -----------------------------------------
# Now we run an Rayleigh Tests to determine if frequency component is significantly phase-locked


pipeline['rayleigh_tests'] = PhaseLockingValue(pipeline['dss_time_epochs_with_gfp'],
                                               n_fft=int(epoch_length*fs*1),
                                               test_frequencies=np.append(random_frequencies,
                                                                          [assr_frequency, 2 * assr_frequency]),
                                               weight_data=True)

# %%
# Generate Phase-locking figures
# ---------------------------------
# Now we plot PLV in the frequency-domain


pipeline['topographic_map_1'] = PlotTopographicMap(pipeline['rayleigh_tests'],
                                                   topographic_channels=np.array(['O2', 'T8', 'T7', 'GFP']),
                                                   plot_x_lim=[0, 90],
                                                   plot_y_lim=[0, 1.2],
                                                   return_figures=True)
pipeline.run()

# %%
# Save results to a database
# ---------------------------------
# We get the measurements we are interested in and save them into a database
subject_info = SubjectInformation(subject_id='Test_Subject')
measurement_info = MeasurementInformation(
    date='Today',
    experiment='sim')

_parameters = {'Type': 'ASSR'}
database_path = reader.input_node.paths.file_directory + os.sep + 'assr_rayleigh_test_data.sqlite'
pipeline['database'] = SaveToDatabase(database_path=database_path,
                                      measurement_information=measurement_info,
                                      subject_information=subject_info,
                                      recording_information={'recording_device': 'dummy_device'},
                                      stimuli_information=_parameters,
                                      processes_list=[pipeline['rayleigh_tests']],
                                      include_waveforms=True
                                      )
pipeline.run()

# %%
# Generate pipeline diagram
# ------------------------------------
pipeline.diagram(file_name=reader.output_node.paths.figures_current_dir + 'pipeline.png',
                 return_figure=True,
                 dpi=600)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import setup, find_packages
import os
from importlib.metadata import version
__version__ = version(__package__)


def run_setup():
    _folder = os.path.dirname(os.path.realpath(__file__))
    requirement_path = _folder + os.path.sep + 'requirements.txt'
    install_requires = []
    if os.path.isfile(requirement_path):
        with open(requirement_path) as f:
            install_requires = f.read().splitlines()
    setup(name="peegy",
          install_requires=install_requires,
          setup_requires=['gitpython'],
          version=__version__,
          packages=find_packages(),
          author="Jaime A. Undurraga",
          author_email="jaime.undurraga@gmail.com",
          description="Tools to pipeline bulk analyses of EEG and other modalities.",
          long_description="""
          Set of tools for processing EEG data data using bdf/edf file format. These can be extended to other modalities
          too. The overall goal is to produce easy processing pipelines to perform batch data analyses in a systematic 
          and reproducible manner. 
          This package includes several statistical, visualization, and output tools to generate consistent SQLITE
          databases. 
          """,
          license="MIT",
          url="https://jundurraga.gitlab.io/peegy/",

          package_data={'': ['*.lay', '*.json']},
          include_package_data=True,
          classifiers=[
              'Development Status :: 3 - Alpha',
              'Environment :: Console',
              'Intended Audience :: Science/Research',
              'License :: OSI Approved :: MIT License',
              'Operating System :: MacOS :: MacOS X',
              'Operating System :: Microsoft :: Windows :: Windows 10',
              'Operating System :: POSIX :: Linux',
              'Programming Language :: Python :: 3',
              'Topic :: Scientific/Engineering :: Bio-Informatics'
              ]
          )
    # update_git_hash_version()


def update_git_hash_version():
    """Return version with local version identifier."""
    import git
    repo = git.Repo(search_parent_directories=True)
    sha = repo.head.object.hexsha
    with open('GITHEADHASH', mode='w+') as f:
        f.write(sha)


if __name__ == '__main__':
    run_setup()



# -*- coding: utf-8 -*-
from peegy.io.eeg.reader import eeg_reader
from peegy.io.edf_bdf_writer import change_subject_id
from peegy.io.external_tools.file_tools import get_files_and_meta_files
import os
from pathlib import Path

_path = Path(os.path.abspath(os.path.dirname(__file__)))
data_folder = _path.parent.parent.absolute().joinpath("test_data/")
data_folder = r'/home/jundurraga/Documents/Measurements/ITD-FR-Pilots/JU/'
data_files_meta = get_files_and_meta_files(data_folder, file_types=['bdf'])
_first_file = data_files_meta.iloc[0].data_links.data_file
_reader = eeg_reader(file_name=_first_file)
print(_reader._header['subject_id'])
change_subject_id(file_name=_first_file, subject_id='S01')
_reader = eeg_reader(file_name=_first_file)
print(_reader._header['subject_id'])



import os.path
from peegy.processing.tools.filters.resampling import eeg_resampling
from peegy.processing.events.event_tools import get_event_times, events_to_samples_array
import os
import numpy as np
import logging
import astropy.units as u
from tqdm import tqdm
import mmap
import datetime
from peegy.processing.events import event_tools
from peegy.io.eeg.reader import eeg_reader
from pathlib import Path
import shutil
log = logging.getLogger()


def data_to_bdf(data: np.array = None,
                events: np.array = None,
                output_file_name: str = None,
                header: dict = None,
                fs: u.Quantity = None):

    n_records = np.ceil((data.shape[0] / fs.to('Hz')).value /
                        header['duration_data_record'].to('s').value).astype(int)
    n_samples_record = np.round(fs.to('Hz').value * header['duration_data_record'].to('s').value).astype(int)

    with open(output_file_name, 'wb') as f:
        # write header (256 bytes)
        f.write(np.uint8(header['identification_code']))
        f.write('{:<7}'.format(header['version'])[0:7].encode('ascii'))
        f.write('{:<80}'.format(header['subject_id'])[0:80].encode('ascii'))
        f.write('{:<80}'.format(header['recording_id'])[0:80].encode('ascii'))
        f.write('{:<8}'.format(header['start_date'])[0:8].encode('ascii'))
        f.write('{:<8}'.format(header['start_time'])[0:8].encode('ascii'))
        f.write(str(header['bytes_in_header']).zfill(8).encode('ascii'))
        f.write('{:<44}'.format(header['data_format'])[0:44].encode('ascii'))
        f.write(str(n_records).zfill(8).encode('ascii'))
        f.write(str(int(header['duration_data_record'].value)).zfill(8)[0:8].encode('ascii'))
        f.write(str(header['n_channels']).zfill(4).encode('ascii'))
        [f.write('{:<16}'.format(header['channels'][i].label)[0:16].encode('ascii')) for i in
         range(header['n_channels'])]
        [f.write('{:<80}'.format(header['transducer'][i])[0:80].encode('ascii')) for i in
         range(header['n_channels'])]
        [f.write('{:<8}'.format(header['physical_dimension'][i])[0:8].encode('ascii')) for i in
         range(header['n_channels'])]
        [f.write('{:<8}'.format(header['physical_minimum'][i])[0:8].encode('ascii')) for i in
         range(header['n_channels'])]
        [f.write('{:<8}'.format(header['physical_maximum'][i])[0:8].encode('ascii')) for i in
         range(header['n_channels'])]
        [f.write('{:<8}'.format(header['digital_minimum'][i])[0:8].encode('ascii')) for i in
         range(header['n_channels'])]
        [f.write('{:<8}'.format(header['digital_maximum'][i])[0:8].encode('ascii')) for i in
         range(header['n_channels'])]
        [f.write('{:<80}'.format(header['pre_filtering'][i])[0:80].encode('ascii')) for i in
         range(header['n_channels'])]
        [f.write('{:<8}'.format(n_samples_record)[0:8].encode('ascii')) for i in
         range(header['n_channels'])]
        [f.write('{:<32}'.format(header['reserved'][i])[0:32].encode('ascii')) for i in
         range(header['n_channels'])]
        # all channels need to have the same units
        assert np.unique(header['physical_dimension'][:-1]).size == 1
        data_to_pack = data.to(header['physical_dimension'][0]).value
        data_events = events
        data_to_pack = np.hstack([data_to_pack, data_events])
        data_to_pack = (data_to_pack / header['gain']).astype(np.int32)

        current_data = np.zeros((n_samples_record, data_to_pack.shape[1]), dtype=np.int32)
        for _nr in range(n_records):
            current_data[:] = 0
            _ini = _nr * n_samples_record
            _end = np.minimum((_nr + 1) * n_samples_record, data_to_pack.shape[0])
            _c_record_length = _end - _ini
            current_data[0:_c_record_length, :] = data_to_pack[_ini: _end, :]
            flat_data = current_data.reshape(-1, order='F')
            encoded_array = flat_data.astype(dtype='<i4').tobytes()
            mm = mmap.mmap(-1, len(encoded_array))
            mm.write(encoded_array)
            mm.seek(0)
            for _i in tqdm(range(flat_data.shape[0]),
                           desc='writing record {:} to bdf file {:}'.format(_nr, output_file_name)):
                f.write(mm.read(3))  # write 3 bytes
                mm.read(1)  # skip 4th byte


def clip_bdf(ini_time: u.Quantity = 0 * u.s,
             end_time: u.Quantity = np.inf * u.s,
             input_file_name: str = None,
             output_file_name: str = None,
             ):
    """
    This function generates a new bdf file from a given to an end time.
    :param ini_time: initial time to obtain data
    :param end_time: end time to obtain data
    :param input_file_name: full path of bdf to obtain the data from
    :param output_file_name: full path of the new bdf file
    :return:
    """

    reader = eeg_reader(file_name=input_file_name)
    header = reader._header
    ini_time = np.round(ini_time / header['duration_data_record']) * header['duration_data_record']
    duration = end_time - ini_time
    time_format = r'%H.%M.%S'
    date_format = r'%d.%m.%y'
    date = datetime.datetime.strptime(header['start_date'] + '-' + header['start_time'],
                                      date_format + '-' + time_format)
    new_starting_date = date + datetime.timedelta(seconds=ini_time.value)
    new_date = new_starting_date.strftime(date_format)
    new_time = new_starting_date.strftime(time_format)

    if end_time == np.inf * u.s:
        duration = header['number_records'] * header['duration_data_record'] - ini_time
    n_records = int(duration / header['duration_data_record'])
    start_offset = (3 * ini_time.value * np.sum(header['number_samples_per_record'])).astype(np.int64)
    buffer_size = (3 * duration.value * np.sum(header['number_samples_per_record'])).astype(np.int64)
    # skip first 256 bytes + 256 * N channels
    header_off_set = 256 * (1 + header['n_channels'])

    with open(output_file_name, 'wb') as f:
        # write header (256 bytes plus 256 * n_channels)
        f.write(np.uint8(header['identification_code']))
        f.write('{:<7}'.format(header['version'])[0:7].encode('ascii'))
        f.write('{:<80}'.format(header['subject_id'])[0:80].encode('ascii'))
        f.write('{:<80}'.format(header['recording_id'])[0:80].encode('ascii'))
        f.write('{:<8}'.format(new_date)[0:8].encode('ascii'))
        f.write('{:<8}'.format(new_time)[0:8].encode('ascii'))
        f.write(str(header['bytes_in_header']).zfill(8).encode('ascii'))
        f.write('{:<44}'.format(header['data_format'])[0:44].encode('ascii'))
        f.write(str(n_records).zfill(8).encode('ascii'))
        f.write(str(int(header['duration_data_record'].value)).zfill(8)[0:8].encode('ascii'))
        f.write(str(header['n_channels']).zfill(4).encode('ascii'))
        [f.write('{:<16}'.format(header['channels'][i].label)[0:16].encode('ascii')) for i in
         range(header['n_channels'])]
        [f.write('{:<80}'.format(header['transducer'][i])[0:80].encode('ascii')) for i in
         range(header['n_channels'])]
        [f.write('{:<8}'.format(header['physical_dimension'][i])[0:8].encode('ascii')) for i in
         range(header['n_channels'])]
        [f.write('{:<8}'.format(header['physical_minimum'][i])[0:8].encode('ascii')) for i in
         range(header['n_channels'])]
        [f.write('{:<8}'.format(header['physical_maximum'][i])[0:8].encode('ascii')) for i in
         range(header['n_channels'])]
        [f.write('{:<8}'.format(header['digital_minimum'][i])[0:8].encode('ascii')) for i in
         range(header['n_channels'])]
        [f.write('{:<8}'.format(header['digital_maximum'][i])[0:8].encode('ascii')) for i in
         range(header['n_channels'])]
        [f.write('{:<80}'.format(header['pre_filtering'][i])[0:80].encode('ascii')) for i in
         range(header['n_channels'])]
        [f.write('{:<8}'.format(header['number_samples_per_record'][i])[0:8].encode('ascii')) for i in
         range(header['n_channels'])]
        [f.write('{:<32}'.format(header['reserved'][i])[0:32].encode('ascii')) for i in
         range(header['n_channels'])]
        with open(input_file_name, 'rb') as input_f:
            input_f.seek(header_off_set + start_offset)
            f.write(input_f.read(buffer_size))


def split_bdf_by_event_code(
              input_file_name: str = None,
              output_path: str = None,
              output_file_name: str = None,
              event_code: float = None,
              ini_offset: u.Quantity = 0 * u.s,
              end_offset: u.Quantity = None,
              trash_original: bool = False
              ):
    """
    This function generates several bdf files based on the event code used to split the data.
    :param input_file_name: full path of bdf to obtain the data from
    :param output_path: path to save files
    :param output_file_name: name of output files. The base name and a number will be returned for each output file
    :param event_code: event number to be found to split file
    :param ini_offset: initial time offset from found event. Note that records offsets will be rounded to
     duration_data_record
    :param end_offset: end time offset from found event. Note that records offsets will be rounded to
    duration_data_record
    :param trash_original: if true, the original file will be moved to .trash folder in the same directory
    :return:
    """

    reader = eeg_reader(file_name=input_file_name)
    raw_events = reader.get_events()
    events = event_tools.Events(event_tools.detect_events(event_channel=raw_events, fs=reader.fs))
    events.summary()
    times = events.get_events_time(code=event_code)
    if not times.size:
        print('no events with code {:} were found in file {:}'.format(event_code, reader.file_name))
        return
    duration_data_record = reader._header['duration_data_record']
    if end_offset is None:
        end_offset = -duration_data_record
    if output_file_name is None:
        _, _file_name = os.path.split(input_file_name)
        output_file_name = _file_name
    if output_path is None:
        _path, _ = os.path.split(input_file_name)
        output_path = Path(_path)
    ini_time = [np.floor((_t + ini_offset) / duration_data_record) * duration_data_record for _t in times]
    end_time = [np.floor((_t + end_offset) / duration_data_record) * duration_data_record for _t in times[1:]]
    end_time.append(np.inf * u.s)
    for _i, (_ini_time, _end_time) in tqdm(enumerate(zip(ini_time, end_time)),
                                           desc='splitting file {:}'.format(reader.file_name)):
        Path(output_path).mkdir(parents=True, exist_ok=True)
        _name, _ext = os.path.splitext(output_file_name)
        _output = Path.joinpath(output_path, '{:}_{:}{:}'.format(_name, _i + 1, _ext))
        print('saving {:}'.format(_output))
        clip_bdf(input_file_name=input_file_name,
                 ini_time=_ini_time,
                 end_time=_end_time,
                 output_file_name=_output)

    if trash_original:
        _path, _name = os.path.split(input_file_name)
        _trash_path = Path(_path).joinpath('.trash')
        Path(_trash_path).mkdir(parents=True, exist_ok=True)
        shutil.move(input_file_name, _trash_path.joinpath(_name))
        print('Original file {:} moved to {:}'.format(input_file_name,
                                                      _trash_path.joinpath(_name)))


def change_subject_id(file_name: str = None,
                      subject_id: str = ''):
    """
    This function will change the subject_id in a .bdf or .edf file. Useful to anonymize
    :param file_name: path to file to be modified
    :param subject_id: id to be written in file
    :return:
    """
    with open(file_name, 'rb+') as f:
        # move to subject_id
        f.seek(8)
        f.write('{:<80}'.format(subject_id)[0:80].encode('ascii'))


def bdf_resampler(
              input_file_name: str = None,
              output_path: str = None,
              output_file_name: str = None,
              trash_original: bool = False,
              new_fs: u.Unit = None
              ):

    if output_file_name is None:
        _, _file_name = os.path.split(input_file_name)
        output_file_name = _file_name
    if output_path is None:
        _path, _ = os.path.split(input_file_name)
        output_path = Path(_path)
    data = eeg_reader(file_name=input_file_name)
    print('gathering data from {:}'.format(input_file_name))
    data_original, events_original, units_orignal, annotations_original = data.get_data()
    data_resampled, _factor = eeg_resampling(x=data_original,
                                             new_fs=new_fs,
                                             fs=data.fs)
    et = get_event_times(event_channel=events_original, fs=data.fs)
    events_resampled = events_to_samples_array(events=et, fs=new_fs, n_samples=data_resampled.shape[0])
    Path(output_path).mkdir(parents=True, exist_ok=True)
    _name, _ext = os.path.splitext(output_file_name)
    _output = Path.joinpath(output_path, '{:}_{:}{:}'.format(_name, 'resampled', _ext))

    data_to_bdf(output_file_name=_output,
                data=data_resampled,
                events=events_resampled,
                header=data._header,
                fs=new_fs)

    if trash_original:
        _path, _name = os.path.split(input_file_name)
        _trash_path = Path(_path).joinpath('.trash')
        Path(_trash_path).mkdir(parents=True, exist_ok=True)
        shutil.move(input_file_name, _trash_path.joinpath(_name))
        print('Original file {:} moved to {:}'.format(input_file_name,
                                                      _trash_path.joinpath(_name)))
